"""
A simple VTK widget for wxPython.  Note that wxPython comes
with its own wxVTKRenderWindow in wxPython.core.vtk.  Try both
and see which one works better for you.

Find wxPython info at http://wxPython.org

Based on vtkTkRenderWindget.py
"""

"""
Please see the example at the end of this file.

----------------------------------------
Creation:

wxVTKRenderWindow(parent, ID, stereo=0, [wx keywords]):

You should create a wxPySimpleApp() or some other wx**App
before creating the window.

----------------------------------------
Methods:

Render()
AddRenderer(ren)
GetRenderers()
GetRenderWindow()

----------------------------------------
Methods to override (all take a wxEvent):

OnButtonDown(event)  default: propagate event to Left, Right, Middle
OnLeftDown(event)    default: set _Mode to 'Rotate'
OnRightDown(event)   default: set _Mode to 'Zoom'
OnMiddleDown(event)  default: set _Mode to 'Pan'

OnButtonUp(event)    default: propagate event to L, R, M and unset _Mode
OnLeftUp(event)
OnRightUp(event)
OnMiddleUp(event)

OnMotion(event)      default: call appropriate handler for _Mode

OnEnterWindow(event) default: set focus to this window
OnLeaveWindow(event) default: release focus

OnKeyDown(event)     default: [R]eset, [W]irefreme, [S]olid, [P]ick
OnKeyUp(event)
OnChar(event)

OnSetFocus(event)
OnKillFocus(event)

OnSize(event)
OnMove(event)

OnPaint(event)       default: Render()

----------------------------------------
Protected Members:

_Mode:                 Current mode: 'Rotate', 'Zoom', 'Pan'
_LastX, _LastY:        The (x,y) coordinates of the previous event
_CurrentRenderer:      The renderer that was most recently clicked in
_CurrentCamera:        The camera for the current renderer

"""

# import usual libraries
import math, os, sys
import wx
# noinspection PyUnresolvedReferences
from vtk import vtkCellPicker, vtkProperty, vtkRenderWindow

# a few configuration items, see what works best on your system

# Use wxGLCanvas as base class instead of wxWindow.
# This is sometimes necessary under wxGTK or the image is blank.
# (in wxWindows 2.3.1 and earlier, the GLCanvas had scroll bars)
# try:
#    wx.WX_USE_GL_CANVAS
# except NameError:
if True:
    if wx.Platform=='__WXMSW__':
        WX_USE_GLCANVAS=0
    else:
        WX_USE_GLCANVAS=1
WX_USE_GLCANVAS=0
# Keep capturing mouse after mouse is dragged out of window
# (in wxGTK 2.3.2 there is a bug that keeps this from working,
# but it is only relevant in wxGTK if there are multiple windows)
# try:
#    wx.WX_USE_X_CAPTURE
# except NameError:
if True:
    if wx.Platform=='__WXMSW__':
        WX_USE_X_CAPTURE=1
    else:
        WX_USE_X_CAPTURE=0

# end of configuration items


if WX_USE_GLCANVAS:
    from wx.glcanvas import GLCanvas

    baseClass=GLCanvas
else:
    baseClass=wx.Window

class wxVTKRenderWindow(baseClass):
    """
    A wxRenderWindow for wxPython.
    Use GetRenderWindow() to get the vtkRenderWindow.
    Create with the keyword stereo=1 in order to
    generate a stereo-capable window.
    """

    def __init__(self, parent, ID, *args, **kw):
        # miscellaneous protected variables
        self._CurrentRenderer=None
        self._CurrentCamera=None
        self._CurrentZoom=1.0
        self._CurrentLight=None

        self._ViewportCenterX=0
        self._ViewportCenterY=0

        self._Picker=vtkCellPicker()
        self._PickedActor=None
        self._PickedProperty=vtkProperty()
        self._PickedProperty.SetColor(1, 0, 0)
        self._PrePickedProperty=None

        # these record the previous mouse position
        self._LastX=0
        self._LastY=0

        # the current interaction mode (Rotate, Pan, Zoom, etc)
        self._Mode=None
        self._ActiveButton=None

        # private attributes
        self.__OldFocus=None

        # used by the LOD actors
        self._DesiredUpdateRate=15
        self._StillUpdateRate=0.0001

        # First do special handling of some keywords:
        # stereo, position, size, width, height, style

        stereo=0

        if 'stereo' in kw:
            if kw['stereo']:
                stereo=1
            del kw['stereo']

        position=wx.DefaultPosition

        if 'position' in kw:
            position=kw['position']
            del kw['position']

        size=wx.DefaultSize

        if 'size' in kw:
            size=kw['size']
            del kw['size']

        if 'width' in kw and 'height' in kw:
            size=(kw['width'], kw['height'])
            del kw['width']
            del kw['height']

        # wxWANTS_CHARS says to give us e.g. TAB
        # wxNO_FULL_REPAINT_ON_RESIZE cuts down resize flicker under GTK
        style=wx.WANTS_CHARS | wx.NO_FULL_REPAINT_ON_RESIZE

        if 'style' in kw:
            style=style | kw['style']
            del kw['style']

        # the enclosing frame must be shown under GTK or the windows
        #  don't connect together properly
        l=[]
        p=parent
        while p:  # make a list of all parents
            l.append(p)
            p=p.GetParent()
        l.reverse()  # sort list into descending order
        for p in l:
            p.Show()

        # initialize the wxWindow
        baseClass.__init__(self, parent, ID, position, size, style)

        # create the RenderWindow and initialize it
        self._RenderWindow=vtkRenderWindow()
        try:
            self._RenderWindow.SetSize(size.width, size.height)
        except AttributeError:
            self._RenderWindow.SetSize(size[0], size[1])
        if stereo:
            self._RenderWindow.StereoCapableWindowOn()
            self._RenderWindow.SetStereoTypeToCrystalEyes()

        self.__Created=0
        # Tell the RenderWindow to render inside the wxWindow.
        if self.GetHandle():
            self.__Created=1
            self._RenderWindow.SetWindowInfo(str(self.GetHandle()))

        # refresh window by doing a Render
        self.Bind(wx.EVT_PAINT, self.OnPaint)
        # turn off background erase to reduce flicker
        self.Bind(wx.EVT_ERASE_BACKGROUND, lambda e: None)

        # Bind the events to the event converters
        self.Bind(wx.EVT_RIGHT_DOWN, self._OnButtonDown)
        self.Bind(wx.EVT_LEFT_DOWN, self._OnButtonDown)
        self.Bind(wx.EVT_MIDDLE_DOWN, self._OnButtonDown)
        self.Bind(wx.EVT_RIGHT_UP, self._OnButtonUp)
        self.Bind(wx.EVT_LEFT_UP, self._OnButtonUp)
        self.Bind(wx.EVT_MIDDLE_UP, self._OnButtonUp)
        self.Bind(wx.EVT_MOTION, self.OnMotion)

        self.Bind(wx.EVT_ENTER_WINDOW, self._OnEnterWindow)
        self.Bind(wx.EVT_LEAVE_WINDOW, self._OnLeaveWindow)

        self.Bind(wx.EVT_CHAR, self.OnChar)
        self.Bind(wx.EVT_KEY_DOWN, self.OnKeyDown)
        self.Bind(wx.EVT_KEY_UP, self.OnKeyUp)

        self.Bind(wx.EVT_SIZE, self._OnSize)
        self.Bind(wx.EVT_MOVE, self.OnMove)

        self.Bind(wx.EVT_SET_FOCUS, self.OnSetFocus)
        self.Bind(wx.EVT_KILL_FOCUS, self.OnKillFocus)

    def SetDesiredUpdateRate(self, rate):
        """Mirrors the method with the same name in
        vtkRenderWindowInteractor."""
        self._DesiredUpdateRate=rate

    def GetDesiredUpdateRate(self):
        """Mirrors the method with the same name in
        vtkRenderWindowInteractor."""
        return self._DesiredUpdateRate

    def SetStillUpdateRate(self, rate):
        """Mirrors the method with the same name in
        vtkRenderWindowInteractor."""
        self._StillUpdateRate=rate

    def GetStillUpdateRate(self):
        """Mirrors the method with the same name in
        vtkRenderWindowInteractor."""
        return self._StillUpdateRate

    def OnPaint(self, event):
        dc=wx.PaintDC(self)
        self.Render()

    def _OnSize(self, event):
        if wx.Platform!='__WXMSW__':
            try:
                width, height=event.GetSize()
            except:
                width=event.GetSize().width
                height=event.GetSize().height
            self._RenderWindow.SetSize(width, height)
        self.OnSize(event)
        if self.__Created:
            self._RenderWindow.Render()

    def OnSize(self, event):
        pass

    def OnMove(self, event):
        pass

    def _OnEnterWindow(self, event):
        self.UpdateRenderer(event)
        self.OnEnterWindow(event)

    def OnEnterWindow(self, event):
        if self.__OldFocus is None:
            self.__OldFocus=wx.Window.FindFocus()
            self.SetFocus()

    def _OnLeaveWindow(self, event):
        self.OnLeaveWindow(event)

    def OnLeaveWindow(self, event):
        if self.__OldFocus:
            self.__OldFocus.SetFocus()
            self.__OldFocus=None

    def OnSetFocus(self, event):
        pass

    def OnKillFocus(self, event):
        pass

    def _OnButtonDown(self, event):
        # helper function for capturing mouse until button released
        self._RenderWindow.SetDesiredUpdateRate(self._DesiredUpdateRate)

        if event.RightDown():
            button="Right"
        elif event.LeftDown():
            button="Left"
        elif event.MiddleDown():
            button="Middle"
        else:
            button=None

        # save the button and capture mouse until the button is released
        if button and not self._ActiveButton:
            self._ActiveButton=button
            if WX_USE_X_CAPTURE:
                self.CaptureMouse()

        self.OnButtonDown(event)

    def OnButtonDown(self, event):
        if not self._Mode:
            # figure out what renderer the mouse is over
            self.UpdateRenderer(event)

        if event.LeftDown():
            self.OnLeftDown(event)
        elif event.RightDown():
            self.OnRightDown(event)
        elif event.MiddleDown():
            self.OnMiddleDown(event)

    def OnLeftDown(self, event):
        if not self._Mode:
            if event.ControlDown():
                self._Mode="Zoom"
            elif event.ShiftDown():
                self._Mode="Pan"
            else:
                self._Mode="Rotate"

    def OnRightDown(self, event):
        if not self._Mode:
            self._Mode="Zoom"

    def OnMiddleDown(self, event):
        if not self._Mode:
            self._Mode="Pan"

    def _OnButtonUp(self, event):
        # helper function for releasing mouse capture
        self._RenderWindow.SetDesiredUpdateRate(self._StillUpdateRate)

        if event.RightUp():
            button="Right"
        elif event.LeftUp():
            button="Left"
        elif event.MiddleUp():
            button="Middle"
        else:
            button=None

        # if the ActiveButton is realeased, then release mouse capture
        if self._ActiveButton and button==self._ActiveButton:
            if WX_USE_X_CAPTURE:
                self.ReleaseMouse()
            self._ActiveButton=None

        self.OnButtonUp(event)

    def OnButtonUp(self, event):
        if event.LeftUp():
            self.OnLeftUp(event)
        elif event.RightUp():
            self.OnRightUp(event)
        elif event.MiddleUp():
            self.OnMiddleUp(event)

        # if not interacting, then do nothing more
        if self._Mode:
            if self._CurrentRenderer:
                self.Render()

        self._Mode=None

    def OnLeftUp(self, event):
        pass

    def OnRightUp(self, event):
        pass

    def OnMiddleUp(self, event):
        pass

    def OnMotion(self, event):
        if self._Mode=="Pan":
            self.Pan(event)
        elif self._Mode=="Rotate":
            self.Rotate(event)
        elif self._Mode=="Zoom":
            self.Zoom(event)

    def OnChar(self, event):
        pass

    def OnKeyDown(self, event):
        if event.GetKeyCode()==ord('R'):
            self.Reset(event)
        if event.GetKeyCode()==ord('W'):
            self.Wireframe()
        if event.GetKeyCode()==ord('S'):
            self.Surface()
        if event.GetKeyCode()==ord('P'):
            self.PickActor(event)

        if event.GetKeyCode()<256:
            self.OnChar(event)

    def OnKeyUp(self, event):
        pass

    def GetZoomFactor(self):
        return self._CurrentZoom

    def GetRenderWindow(self):
        return self._RenderWindow

    def GetPicker(self):
        return self._Picker

    def Render(self):
        if self._CurrentLight:
            light=self._CurrentLight
            light.SetPosition(self._CurrentCamera.GetPosition())
            light.SetFocalPoint(self._CurrentCamera.GetFocalPoint())

        if self.__Created:
            self._RenderWindow.Render()
        elif self.GetHandle():
            self._RenderWindow.SetWindowInfo(str(self.GetHandle()))
            self.__Created=1
            self._RenderWindow.Render()

    def UpdateRenderer(self, event):
        """
        UpdateRenderer will identify the renderer under the mouse and set
        up _CurrentRenderer, _CurrentCamera, and _CurrentLight.
        """
        x=event.GetX()
        y=event.GetY()
        windowX, windowY=self._RenderWindow.GetSize()

        renderers=self._RenderWindow.GetRenderers()
        numRenderers=renderers.GetNumberOfItems()

        self._CurrentRenderer=None
        renderers.InitTraversal()
        for i in range(0, numRenderers):
            renderer=renderers.GetNextItem()
            vx, vy=(0, 0)
            if windowX>1:
                vx=float(x)/(windowX-1)
            if windowY>1:
                vy=(windowY-float(y)-1)/(windowY-1)
            (vpxmin, vpymin, vpxmax, vpymax)=renderer.GetViewport()

            if vpxmin<=vx<=vpxmax and vpymin<=vy<=vpymax:
                self._CurrentRenderer=renderer
                self._ViewportCenterX=float(windowX)*(vpxmax-vpxmin)/2.0 \
                                      +vpxmin
                self._ViewportCenterY=float(windowY)*(vpymax-vpymin)/2.0 \
                                      +vpymin
                self._CurrentCamera=self._CurrentRenderer.GetActiveCamera()
                lights=self._CurrentRenderer.GetLights()
                lights.InitTraversal()
                self._CurrentLight=lights.GetNextItem()
                break

        self._LastX=x
        self._LastY=y

    def GetCurrentRenderer(self):
        return self._CurrentRenderer

    def Rotate(self, event):
        if self._CurrentRenderer:
            x=event.GetX()
            y=event.GetY()

            self._CurrentCamera.Azimuth(self._LastX-x)
            self._CurrentCamera.Elevation(y-self._LastY)
            self._CurrentCamera.OrthogonalizeViewUp()

            self._LastX=x
            self._LastY=y

            self._CurrentRenderer.ResetCameraClippingRange()
            self.Render()

    def Pan(self, event):
        if self._CurrentRenderer:
            x=event.GetX()
            y=event.GetY()

            renderer=self._CurrentRenderer
            camera=self._CurrentCamera
            (pPoint0, pPoint1, pPoint2)=camera.GetPosition()
            (fPoint0, fPoint1, fPoint2)=camera.GetFocalPoint()

            if camera.GetParallelProjection():
                renderer.SetWorldPoint(fPoint0, fPoint1, fPoint2, 1.0)
                renderer.WorldToDisplay()
                fx, fy, fz=renderer.GetDisplayPoint()
                renderer.SetDisplayPoint(fx-x+self._LastX,
                                         fy+y-self._LastY,
                                         fz)
                renderer.DisplayToWorld()
                fx, fy, fz, fw=renderer.GetWorldPoint()
                camera.SetFocalPoint(fx, fy, fz)

                renderer.SetWorldPoint(pPoint0, pPoint1, pPoint2, 1.0)
                renderer.WorldToDisplay()
                fx, fy, fz=renderer.GetDisplayPoint()
                renderer.SetDisplayPoint(fx-x+self._LastX,
                                         fy+y-self._LastY,
                                         fz)
                renderer.DisplayToWorld()
                fx, fy, fz, fw=renderer.GetWorldPoint()
                camera.SetPosition(fx, fy, fz)

            else:
                (fPoint0, fPoint1, fPoint2)=camera.GetFocalPoint()
                # Specify a point location in world coordinates
                renderer.SetWorldPoint(fPoint0, fPoint1, fPoint2, 1.0)
                renderer.WorldToDisplay()
                # Convert world point coordinates to display coordinates
                dPoint=renderer.GetDisplayPoint()
                focalDepth=dPoint[2]

                aPoint0=self._ViewportCenterX+(x-self._LastX)
                aPoint1=self._ViewportCenterY-(y-self._LastY)

                renderer.SetDisplayPoint(aPoint0, aPoint1, focalDepth)
                renderer.DisplayToWorld()

                (rPoint0, rPoint1, rPoint2, rPoint3)=renderer.GetWorldPoint()
                if rPoint3!=0.0:
                    rPoint0=rPoint0/rPoint3
                    rPoint1=rPoint1/rPoint3
                    rPoint2=rPoint2/rPoint3

                camera.SetFocalPoint((fPoint0-rPoint0)+fPoint0,
                                     (fPoint1-rPoint1)+fPoint1,
                                     (fPoint2-rPoint2)+fPoint2)

                camera.SetPosition((fPoint0-rPoint0)+pPoint0,
                                   (fPoint1-rPoint1)+pPoint1,
                                   (fPoint2-rPoint2)+pPoint2)

            self._LastX=x
            self._LastY=y

            self.Render()

    def Zoom(self, event):
        if self._CurrentRenderer:
            x=event.GetX()
            y=event.GetY()

            renderer=self._CurrentRenderer
            camera=self._CurrentCamera

            zoomFactor=math.pow(1.02, (0.5*(self._LastY-y)))
            self._CurrentZoom=self._CurrentZoom*zoomFactor

            if camera.GetParallelProjection():
                parallelScale=camera.GetParallelScale()/zoomFactor
                camera.SetParallelScale(parallelScale)
            else:
                camera.Dolly(zoomFactor)
                renderer.ResetCameraClippingRange()

            self._LastX=x
            self._LastY=y

            self.Render()

    def Reset(self, event=None):
        if self._CurrentRenderer:
            self._CurrentRenderer.ResetCamera()

        self.Render()

    def Wireframe(self):
        actors=self._CurrentRenderer.GetActors()
        numActors=actors.GetNumberOfItems()
        actors.InitTraversal()
        for i in range(0, numActors):
            actor=actors.GetNextItem()
            actor.GetProperty().SetRepresentationToWireframe()

        self.Render()

    def Surface(self):
        actors=self._CurrentRenderer.GetActors()
        numActors=actors.GetNumberOfItems()
        actors.InitTraversal()
        for i in range(0, numActors):
            actor=actors.GetNextItem()
            actor.GetProperty().SetRepresentationToSurface()

        self.Render()

    def PickActor(self, event):
        if self._CurrentRenderer:
            x=event.GetX()
            y=event.GetY()

            renderer=self._CurrentRenderer
            picker=self._Picker

            windowX, windowY=self._RenderWindow.GetSize()
            picker.Pick(x, (windowY-y-1), 0.0, renderer)
            actor=picker.GetActor()

            if (self._PickedActor is not None and
                    self._PrePickedProperty is not None):
                self._PickedActor.SetProperty(self._PrePickedProperty)
                # release hold of the property
                self._PrePickedProperty.UnRegister(self._PrePickedProperty)
                self._PrePickedProperty=None

            if actor is not None:
                self._PickedActor=actor
                self._PrePickedProperty=self._PickedActor.GetProperty()
                # hold onto the property
                self._PrePickedProperty.Register(self._PrePickedProperty)
                self._PickedActor.SetProperty(self._PickedProperty)

            self.Render()




'''
=========
UserFuncs
=========

A plugin so that users can include their own user function into
the model script and execute them inside the GUI of GenX.<p>

The following design criteria exists:
The function has to be defined inside the script.
The function has to take zero input arguments.

Thus the following example can serve as a template:
::
    def myuserfunc():
        # Do something
        print('It works!')

This should be added somewhere in the script. This provides a great way
to, for example, dump some internal data into a file or for checking the status
of some variables.

For more complex actions you can use type annotation to allow the Plugin
to automatically build a dialog for the user to enter values.
This is implemented for int, float and str types.
If a variable is annotated with **TextIO** a subsequend file
selection dialog is shown to the user to select a file to save to.
The open file object is then passed on to the function:
::
    def my_func_params(a: int, b: float, c: float=1.0, name: str="Name", file_handler: TextIO=None):
        file_handler.write(f'a={a};b={b};c={c};name={name}')

This gives you the flexibility to perform complex tasks while still benefitting from
a GUI integration. Any number of parameters or files can be defined for your funct.
If a parameter is given as keywords, the default dialog value is taken from
the keyword default, otherwise it is 0 for numbers or an empty string.
The default for a TextIO object is ignored.
'''
import types
import typing
import wx
import io
import traceback
import inspect
from contextlib import ExitStack

from .. import add_on_framework as framework
from .help_modules.custom_dialog import ValueValidator, ValidateDialog


DEFAULT_VALUE = {int: 0, float: 0., str: ''}


class Plugin(framework.Template):
    userfunc_arguments: dict

    def __init__(self, parent):
        framework.Template.__init__(self, parent)
        self.menu = self.NewMenu('User funcs')
        self.userfunc_arguments = {}
        self.parent = parent
        self.StatusMessage('Sucessfully loaded UserFuncs...')

    def OnSimulate(self, event):
        '''OnSimulate(self, event) --> None
        
        Updates stuff after simulation
        '''
        model = self.GetModel()
        # locate all functions in the model.script_module
        funcs = [eval('model.script_module.%s'%name, globals(), {'model': model}) \
                 for name in dir(model.script_module) \
                 if type(eval('model.script_module.%s'%name, globals(), {'model': model})
                         )==types.FunctionType]
        # Find the functions that are defined in the script and take zero
        # input parameters
        user_funcs = [f.__name__ for f in funcs if \
                      f.__module__=='genx_script_module' \
                      and f.__code__.co_argcount==0]

        # Find the functions that are defined in the script and have type annotation
        # with sufficient information to create and automatic dialog
        self.userfunc_arguments = {}
        for func in [f for f in funcs if f.__module__=='genx_script_module' and f.__code__.co_argcount!=0]:
            pars = dict(inspect.signature(func).parameters)
            args = pars.keys()
            defaults = dict([(n, p.default) for n,p in pars.items()])

            annots = getattr(func, '__annotations__')
            args_ok = True
            for ai in args:
                ani = annots.get(ai)
                if not ani in [int, float, str, typing.TextIO]:
                    args_ok = False
                    break
            if args_ok:
                name = func.__name__
                self.userfunc_arguments[name] = (args, annots, defaults)
                user_funcs.append(name)

        # Remove all the previous functions
        self.clear_menu()
        # Lets add our user functions to our custo menu
        for name in user_funcs:
            menuitem = wx.MenuItem(self.menu, wx.NewId(), name,
                                   'Evaluate %s'%name, wx.ITEM_NORMAL)
            self.menu.Append(menuitem)
            self.parent.Bind(wx.EVT_MENU, self.eh_menu_choice, menuitem)

    def clear_menu(self):
        '''clear_menu(self) --> None
        
        Clears the menu from all items present in it
        '''
        [self.menu.Remove(item) for item in self.menu.GetMenuItems()]

    def eh_menu_choice(self, event):
        '''eh_menu_choice(self, event)
        
        event handler for the choice of an menuitem in the User Functions 
        menu. Executes the function as defined in the script. 
        With error catching.
        '''
        menu = event.GetEventObject()
        menuitem = menu.FindItemById(event.Id)
        fname = menuitem.GetItemLabelText()
        # Now try to evaluate the function
        self.StatusMessage('Trying to evaluate %s'%fname)
        if fname in self.userfunc_arguments:
            self.exec_args_function(fname)
        else:
            self.exec_simple_function(fname)

    def exec_simple_function(self, fname):
        model = self.GetModel()
        try:
            model.eval_in_model(f'{fname}()')
        except Exception as e:
            # a bit of traceback
            outp = io.StringIO()
            traceback.print_exc(200, outp)
            tb_string = outp.getvalue()
            outp.close()
            self.ShowWarningDialog('Error in evaluating the' \
                                   ' function: %s.\n The error gave the following traceback:' \
                                   '\n%s'%(fname, tb_string))
            self.StatusMessage('error in function %s'%fname)
        else:
            self.StatusMessage('%s sucessfully evaluated'%fname)

    def get_filename(self, argname):
        dia = wx.FileDialog(self.parent, f'Select file for {argname}', style=wx.FD_SAVE| wx.FD_OVERWRITE_PROMPT)
        dia.ShowModal()
        fname = dia.GetPath()
        dia.Destroy()
        return fname

    def exec_args_function(self, fname):
        model = self.GetModel()
        args, annots, defaults = self.userfunc_arguments[fname]
        dia_args = [ai for ai in args if annots[ai] is not typing.TextIO]
        dia_vals = dict([(ai, defaults[ai]) if defaults[ai] is not inspect.Parameter.empty
                        else (ai, DEFAULT_VALUE[annots[ai]])
                        for ai in dia_args])
        dia_validate = dict([(ai, ValueValidator(annots[ai])) for ai in dia_args])

        file_args = [ai for ai in args if annots[ai] is typing.TextIO]

        if len(dia_args)>0:
            dia = ValidateDialog(self.parent, dia_args, dia_vals, dia_validate, title=f'Evaluate parameters for {fname}')
            res = dia.ShowModal()
            call_values = dia.GetValues()
            dia.Destroy()
            if res!=wx.ID_OK:
                return
            for key, value in call_values.items():
                # convert entry value into correct type
                call_values[key] = annots[key](value)
        else:
            call_values = {}

        with ExitStack() as stack:
            for fi in file_args:
                call_values[fi] = stack.enter_context(open(self.get_filename(fi), 'w', encoding='utf8'))

            call_pars = [call_values[ai] for ai in args]
            model.set_script_variable('_call_pars', call_pars)

            try:
                model.eval_in_model(f'{fname}(*_call_pars)')
            except Exception as e:
                # a bit of traceback
                outp = io.StringIO()
                traceback.print_exc(200, outp)
                tb_string = outp.getvalue()
                outp.close()
                self.ShowWarningDialog('Error in evaluating the' \
                                       ' function: %s.\n The error gave the following traceback:' \
                                       '\n%s'%(fname, tb_string))
                self.StatusMessage('error in function %s'%fname)
            else:
                self.StatusMessage('%s sucessfully evaluated'%fname)
            finally:
                model.unset_script_variable('_call_pars')



# -*- coding: utf8 -*-
'''
=============
SpinAsymmetry
=============

A plugin to show an additional graph to plot the spin-asymmetry of polarized measurements.

Written by Artur Glavic
'''

import wx
from numpy import isnan
from genx.gui.plotpanel import PlotPanel, BasePlotConfig
from .. import add_on_framework as framework

class SAPlotConfig(BasePlotConfig):
    section='spin asymmetry plot'

class SAPlotPanel(wx.Panel):
    ''' Widget for plotting the spin-asymmetry of datasets.
    '''

    def __init__(self, parent, plugin, id=-1, color=None, dpi=None
                 , style=wx.NO_FULL_REPAINT_ON_RESIZE, **kwargs):
        wx.Panel.__init__(self, parent)
        self.plot=PlotPanel(self, -1, color, dpi, SAPlotConfig, style, **kwargs)
        self.plugin=plugin

        sizer=wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.plot, 1, wx.EXPAND | wx.GROW | wx.ALL)

        self.plot.update(None)
        self.plot.ax=self.plot.figure.add_subplot(111)
        box=self.plot.ax.get_position()
        self.plot.ax.set_position([box.x0, box.y0, box.width*0.95, box.height])
        self.plot.ax.set_autoscale_on(True)
        self.plot.update=self.Plot
        self.SetSizer(sizer)
        self.plot_dict={}

    def SetZoom(self, active=False):
        return self.plot.SetZoom(active)

    def GetZoom(self):
        return self.plot.GetZoom()

    def Plot(self):
        ''' Plot(self) --> None

        Plotting the sample Sample.
        '''
        colors=['b', 'r', 'g', 'c', 'm', 'y', 'k']
        # self.plot_dict = model.sample.SimSLD(None, model.inst)
        while len(self.plot.ax.lines)>0:
            self.plot.ax.lines[0].remove()
        i=0
        data=self.plugin.GetModel().get_data()
        sld_units=[]

        # New style sim function with one sld for each simulation
        for i, di in enumerate(data):
            if i%2!=0:
                continue
            if di.show:
                try:
                    dj=data[i+1]
                    li=len(di.x)
                    lj=len(dj.x)
                    l=min(li, lj)
                    SAdata=(di.y[:l]-dj.y[:l])/(di.y[:l]+dj.y[:l])
                    if not isnan(SAdata).all():
                        self.plot.ax.plot(di.x[:l], SAdata,
                                          color=di.data_color,
                                          ls=di.data_linetype, lw=di.data_linethickness,
                                          marker=di.data_symbol, ms=di.data_symbolsize,
                                          label='data '+di.name+'&'+dj.name)
                    SAsim=(di.y_sim[:l]-dj.y_sim[:l])/(di.y_sim[:l]+dj.y_sim[:l])
                    if not isnan(SAsim).all():
                        self.plot.ax.plot(di.x[:l], SAsim,
                                          color=di.sim_color,
                                          ls=di.sim_linetype, lw=di.sim_linethickness,
                                          marker=di.sim_symbol, ms=di.sim_symbolsize,
                                          label='sim '+di.name+'&'+dj.name)
                except:
                    pass

        self.plot.ax.legend(loc='upper right',  # bbox_to_anchor=(1, 0.5),
                            framealpha=0.5,
                            fontsize="small", ncol=1)

        self.plot.ax.yaxis.label.set_text('Spin Asymmetry')
        self.plot.ax.xaxis.label.set_text('x')
        wx.CallAfter(self.plot.flush_plot)
        self.plot.AutoScale()

    def ReadConfig(self):
        '''ReadConfig(self) --> None

        Reads in the config file
        '''
        return self.plot.ReadConfig()

    def GetYScale(self):
        '''GetYScale(self) --> String

        Returns the current y-scale in use. Currently the string
        'log' or 'linear'. If the axes does not exist it returns None.
        '''
        if self.plot.ax:
            return self.plot.ax.get_yscale()
        else:
            return None

    def GetXScale(self):
        '''GetXScale(self) --> String

        Returns the current x-scale in use. Currently the string
        'log' or 'linear'. If the axes does not exist it returns None.
        '''
        if self.plot.ax:
            return self.plot.ax.get_xscale()
        else:
            return None

class Plugin(framework.Template):
    def __init__(self, parent):
        framework.Template.__init__(self, parent)
        self.parent=parent

        # Create the SLD plot
        SA_plot_panel=self.NewPlotFolder('Spin-Asymmetry')
        SA_sizer=wx.BoxSizer(wx.HORIZONTAL)
        SA_plot_panel.SetSizer(SA_sizer)
        self.SA_plot=SAPlotPanel(SA_plot_panel, self)
        SA_sizer.Add(self.SA_plot, 1, wx.EXPAND | wx.GROW | wx.ALL)
        SA_plot_panel.Layout()

    def OnSimulate(self, event):
        '''OnSimulate(self, event) --> None

        Updates stuff after simulation
        '''
        # Calculate and update the sld plot
        wx.CallAfter(self.SA_plot.Plot)

    def OnFittingUpdate(self, event):
        '''OnSimulate(self, event) --> None

        Updates stuff during fitting
        '''
        wx.CallAfter(self.SA_plot.Plot())



"""
Executable module to start a server that can be used for
remote refinement on a server.
"""
import os
import asyncio
import argparse
import appdirs
import logging

from . import version
from .core import custom_logging


def set_numba_single():
    config_path = os.path.abspath(appdirs.user_data_dir('GenX3', 'ArturGlavic'))
    cache_dir = os.path.join(config_path, 'single_cpu_numba_cache')

    logging.debug('Setting numba JIT compilation to single CPU')
    import numba
    numba.config.CACHE_DIR = cache_dir
    old_jit = numba.jit

    def jit(*args, **opts):
        opts['parallel'] = False
        return old_jit(*args, **opts)

    numba.jit = jit
    numba.GENX_OVERWRITE_SINGLE = True
    try:
        numba.set_num_threads(1)
    except AttributeError:
        pass


def main():
    import multiprocessing
    multiprocessing.set_start_method('spawn')
    try:
        from mpi4py import MPI
    except ImportError:
        rank = 0
        __mpi__ = False
    else:
        __mpi__ = bool(MPI.COMM_WORLD.Get_size()>1)
        rank = MPI.COMM_WORLD.Get_rank()

    parser = argparse.ArgumentParser(description="GenX %s, fits data to a model. Server script for remote fitting."%version.__version__,
                                     epilog="For support, manuals and bug reporting see http://genx.sf.net"
                                     )
    parser.add_argument('address', nargs='?', default='localhost', help='Network address to listen on')
    parser.add_argument('port', nargs='?', default=3000, help='Network port to listen on')
    parser.add_argument('-l', '--logfile', dest='logfile', default=None, type=str,
                        help='Output debug information to logfile.')
    parser.add_argument('-p', '--password', dest='password', default='empty', type=str,
                        help='Password used when establishing connection from client.')
    parser.add_argument('--debug', dest='debug', default=False, action="store_true",
                        help='Show additional debug information on console/logfile')
    parser.add_argument('--disable-nb', dest='disable_numba', default=False, action="store_true",
                        help='Disable the use of numba JIT compiler')
    parser.add_argument('--no-nb1', dest='numba_single', default=True, action="store_false",
                        help='Compile numba JIT functions with parallel computing support (use more then one core). '
                             'This is the default for no-server based fits.')

    args = parser.parse_args()

    if not args.debug:
        custom_logging.CONSOLE_LEVEL = logging.INFO
    if rank>0:
        custom_logging.CONSOLE_LEVEL = logging.WARNING
    custom_logging.setup_system()

    if args.logfile:
        if __mpi__:
            lfbase, lfend = args.logfile.rsplit('.', 1)
            custom_logging.activate_logging(f'{lfbase}_{rank:02}.{lfend}')
        else:
            custom_logging.activate_logging(args.logfile)
    if rank==0:
        logging.debug("Arguments from parser: %s"%args)

    if args.disable_numba:
        from genx.models import lib as modellib
        modellib.USE_NUMBA = False
    else:
        if args.numba_single:
            try:
                set_numba_single()
            except ImportError:
                pass
        if rank==0:
            logging.info('Importing numba based modules to pre-compile JIT functions, this can take some time')
        from genx.models.lib import paratt_numba, neutron_numba, instrument_numba, offspec, surface_scattering
        if rank==0:
            logging.info('Modules imported successfully')

    if rank==0:
        logging.info('Starting RemoteController')
    from .remote import controller
    ctrl = controller.RemoteController()
    ctrl.key = args.password.encode('utf-8')
    asyncio.run(ctrl.serve(args.address, args.port))


if __name__=='__main__':
    main()



'''
Module to handle plugins including a template class for writing
plugins. 
'''

import os
import wx, io, traceback
from dataclasses import dataclass
from logging import debug, info, error, warning

from genx.core.config import BaseConfig, Configurable
from .utils import PluginHandler, ShowInfoDialog, ShowErrorDialog, ShowWarningDialog, ShowQuestionDialog

head, tail=os.path.split(__file__)
# Look only after the file name and not the ending since
# the file ending can be pyc if compiled... 
__FILENAME__=tail.split('.')[0]
# This assumes that plugin is under the current dir may need 
# changing
__MODULE_DIR__=head
if __MODULE_DIR__!='/':
    __MODULE_DIR__+='/'

class Template:
    ''' A template class for handling plugins. Note that using the 
    New* methods will automatically remove them when removing the plugin.
    Otherwise the programmer, if he/she makes more advanced changes in the gui,
    have to take care of the deletion of objects.
    '''

    def __init__(self, parent):
        '''
        This method should be overloaded.
        '''
        self.parent=parent
        self.plot_pages=[]
        self.input_pages=[]
        self.data_pages=[]
        self.menus=[]

    def InputPageChanged(self, pname):
        # Can be used to react to own plugin pages being selected
        pass

    def NewPlotFolder(self, name, pos=-1):
        '''
        Creates a new Folder in the Plot part of the panels. Returns
        a wx.Panel which can be used to create custom controls. 
        o not forget to use RemovePlotfolder in the Destroy method.
        '''
        panel=wx.Panel(self.parent.plot_notebook, -1)
        self.parent.plot_notebook.AddPage(panel, name)
        self.plot_pages.append(panel)

        return panel

    def NewInputFolder(self, name, pos=-1):
        '''
        Creates a new Folder in the Input part of the panels. Returns
        a wx.Panel which can be used to create custom controls. 
        o not forget to use RemoveInputfolder in the Destroy method.
        '''
        panel=wx.Panel(self.parent.input_notebook, -1)
        self.parent.input_notebook.AddPage(panel, name)
        self.input_pages.append(panel)

        return panel

    def NewDataFolder(self, name, pos=-1):
        '''
        Creates a new Folder in the data part of the panels. Returns
        a wx.Panel which can be used to create custom controls. 
        o not forget to use RemoveInputfolder in the Destroy method.
        '''
        panel=wx.Panel(self.parent.data_notebook, -1)
        self.parent.data_notebook.AddPage(panel, name)
        self.data_pages.append(panel)

        return panel

    def NewMenu(self, name):
        '''
        Creates an top menu that can be used to control the plugin. Remeber
        to also implement RemoveMenu in the Destroy method.
        '''
        menu=wx.Menu()
        self.parent.main_frame_menubar.Append(menu, name)
        self.menus.append(name)

        return menu

    def StatusMessage(self, text):
        '''
        Method that sets the staustext in the main window
        '''
        debug(text)
        self.parent.main_frame_statusbar.SetStatusText(text, 1)

    def ShowErrorDialog(self, message):
        '''
        Shows an error dialog with message [string]
        '''
        error(message)
        ShowErrorDialog(self.parent, message)

    def ShowInfoDialog(self, message):
        '''
        Shows an info dialog with message [string]
        '''
        info(message)
        ShowInfoDialog(self.parent, message)

    def ShowWarningDialog(self, message):
        '''
        Shows an warning dialog with message [string]
        '''
        warning(message)
        ShowWarningDialog(self.parent, message)

    def ShowQuestionDialog(self, message):
        '''
        Shows an warning dialog with message [string]
        '''
        return ShowQuestionDialog(self.parent, message)

    def GetModel(self):
        '''
        Returns the model currently in use. This is a pointer to the model
        object thus it will automatically always conatin the newest information.        
        '''
        return self.parent.model_control.get_model()

    def GetSolverControl(self):
        '''
        Returns the solver_control object that controls all aspects of
        the calculational part of the fitting.
        '''
        return self.parent.model_control

    def SetModelScript(self, script):
        '''
        Sets the script of the current model. This overwrite the current 
        script.
        '''
        self.parent.model_control.set_model_script(script)

    def GetModelScript(self):
        """Returns the model script"""
        return self.parent.model_control.get_model_script()

    def CompileScript(self):
        '''
        Compiles the model script
        '''
        self.parent.model_control.force_compile()

    def GetScriptModule(self):
        """Returns the script module"""
        return self.parent.model_control.script_module

    def OnNewModel(self, event):
        '''
        Function to be overridden. Called when a new model is being created.
        '''
        pass

    def OnDataChanged(self, event):
        '''OnDataChanged(self) --> None
        
        Function to be overridden. Called when a new data set has been loaded
        or deleted.
        '''
        pass

    def OnOpenModel(self, event):
        '''OnOpenModel(self, event) --> None
        
        Function that is called after a new model has been loaded.
        Used to set up plugin specific model stuff. To be overridden
        '''
        pass

    def OnSimulate(self, event):
        '''OnSimulate(self, event) --> None
        
        Function that is called after a simulation has been done.
        To be overridden
        '''
        pass

    def OnFittingUpdate(self, event):
        '''OnFittingUpdate(self, event) --> None
        
        Function that is called when the fitting algorithm pushes an update event.
        To be overridden
        '''
        pass

    def OnGridChange(self, event):
        """Responds to a change in the grid.

        :param event:
        :return:
        """
        pass

    def Remove(self):
        '''
        Removes all components.
        '''
        pnb=self.parent.plot_notebook
        inb=self.parent.input_notebook
        dnb=self.parent.data_notebook

        # remove all pages from the notebooks
        for panel in self.plot_pages:
            pages=[pnb.GetPage(i) for i in range(pnb.GetPageCount())]
            idx=pages.index(panel)
            pnb.DeletePage(idx)
        for panel in self.input_pages:
            pages=[inb.GetPage(i) for i in range(inb.GetPageCount())]
            idx=pages.index(panel)
            inb.DeletePage(idx)
        for panel in self.data_pages:
            pages=[dnb.GetPage(i) for i in range(dnb.GetPageCount())]
            idx=pages.index(panel)
            dnb.DeletePage(idx)
        # Remove the menus
        for name in self.menus:
            idx=self.parent.main_frame_menubar.FindMenu(name)
            self.parent.main_frame_menubar.Remove(idx)

@dataclass
class PluginConfig(BaseConfig):
    section='plugins'
    loaded_plugins: str=''

class PluginController(Configurable):
    '''
    A controller class to interact with the gui
    so we can load and unload modules as well as
    update the module list.
    '''
    opt: PluginConfig

    def __init__(self, parent, menu):
        '''
        Insert menu items for controlling plugins in menu.
        Parent is the main window.
        '''
        self.plugin_handler=PluginHandler(parent, __MODULE_DIR__, 'add_ons')
        self.parent=parent
        Configurable.__init__(self)

        # make the menus
        self.load_menu=wx.Menu()
        menu.Insert(0, -1, 'Load', self.load_menu, 'Load a plugin')
        self.unload_menu=wx.Menu()
        menu.Insert(1, -1, 'Unload', self.unload_menu, 'Unload a plugin')

        menu.Append(-1, 'Update module list')

    def update_plugins(self):
        '''update_modules(self) --> None
        
        Updates the list of modules that can be loaded.
        '''
        # Remove all the items in load_menu
        items=self.load_menu.GetMenuItems()
        [self.load_menu.Delete(item) for item in items]

        # Get the new list of plugin modules
        modlist=self.plugin_handler.get_plugins()
        modlist.sort()

        # Add new menu items
        for mod in modlist:
            menu=self.load_menu.Append(-1, mod)
            self.parent.Bind(wx.EVT_MENU, self.LoadPlugin, menu)

        self.update_config()

    def RegisterPlugin(self, plugin):
        ''' RegisterPlugin(self, plugin) --> None
        
        Adds a plugin to the unload list so that it can be removed later.
        '''
        menu=self.unload_menu.Append(-1, plugin)
        self.parent.Bind(wx.EVT_MENU, self.UnLoadPlugin, menu)
        self.update_plugins()

    def update_config(self):
        '''update_config(self) --> None
        
        Updates the config object
        '''
        loaded_plugins=self.plugin_handler.get_loaded_plugins()
        self.opt.loaded_plugins=';'.join(loaded_plugins)
        self.WriteConfig()

    def LoadDefaultPlugins(self):
        '''LoadDefaultPlugins(self) --> None
        
        Tries to load the default plugins from the config object
        if they are not already loaded.
        '''
        self.ReadConfig()
        plugin_str=self.opt.loaded_plugins

        if plugin_str=='':
            self.update_plugins()
            return
        existing_plugins=self.plugin_handler.get_possible_plugins()

        for plugin in plugin_str.split(';'):
            # Check so the plugin is not loaded and exists 
            if not self.plugin_handler.is_loaded(plugin):
                if plugin in existing_plugins:
                    try:
                        self.plugin_handler.load_plugin(plugin)
                        self.RegisterPlugin(plugin)
                    except:
                        outp=io.StringIO()
                        traceback.print_exc(200, outp)
                        tbtext=outp.getvalue()
                        outp.close()
                        ShowErrorDialog(self.parent, 'Can NOT load plugin ' \
                                        +plugin+'\nPython traceback below:\n\n'+tbtext)
                        self.RegisterPlugin(plugin)
                else:
                    ShowInfoDialog(self.parent, 'Could not find plugin "%s"' \
                                                '. Either there is an error in the config file' \
                                                ' or the plugin is not installed.'%plugin)
        self.update_plugins()

    # Callbacks
    def LoadPlugin(self, event):
        '''OnLoadPlugin(self, event) --> None
        
        Loads a plugin from a menu choice.
        '''
        # Get the name of the plugin
        menuitem=self.load_menu.FindItemById(event.GetId())
        plugin=menuitem.GetItemLabel()
        try:
            self.plugin_handler.load_plugin(plugin)
        except:
            outp=io.StringIO()
            traceback.print_exc(200, outp)
            tbtext=outp.getvalue()
            outp.close()
            ShowErrorDialog(self.parent, 'Can NOT load plugin '+plugin \
                            +'\nPython traceback below:\n\n'+tbtext)
        else:
            self.RegisterPlugin(plugin)

    def UnLoadPlugin(self, event):
        '''UnLoadPlugin(self, event) --> None
        
        UnLoads (removes) a plugin module.
        '''
        menuitem=self.unload_menu.FindItemById(event.GetId())
        plugin=menuitem.GetItemLabel()
        if self.UnLoadPlugin_by_Name(plugin):
            # Remove the item from the list
            self.unload_menu.Delete(menuitem)
            # Update the available plugins
            self.update_plugins()

    def UnLoadPlugin_by_Name(self, plugin):
        try:
            self.plugin_handler.unload_plugin(plugin)
        except Exception as e:
            outp=io.StringIO()
            traceback.print_exc(200, outp)
            tbtext=outp.getvalue()
            outp.close()
            ShowErrorDialog(self.parent, 'Can NOT unload plugin object'+ \
                            plugin+'\nPython traceback below:\n\n'+tbtext)
            return False
        else:
            return True

    def OnNewModel(self, event):
        '''OnNewModel(self, event) --> None
        
        Runs plugin code when the user tries to load a new model 
        '''
        for name in self.plugin_handler.loaded_plugins:
            self.plugin_handler.loaded_plugins[name].OnNewModel(event)

    def OnDataChanged(self, event):
        '''OnNewModel(self, event) --> None
        
        Runs plugin code when the user tries to load new data
        '''
        for name in self.plugin_handler.loaded_plugins:
            self.plugin_handler.loaded_plugins[name].OnDataChanged(event)

    def OnOpenModel(self, event):
        '''OnOpenModel(self, event) --> None
        
        Runs plugin code when the user tries to open a model 
        '''
        loaded_plugins=list(self.plugin_handler.loaded_plugins.keys())
        items=self.unload_menu.GetMenuItems()
        for item in items:
            # Remove the item from the list
            self.unload_menu.Delete(item)

        for name in loaded_plugins:
            # self.plugin_handler.loaded_plugins[name].OnOpenModel(event)
            self.plugin_handler.unload_plugin(name)

        self.LoadDefaultPlugins()
        # Update the available plugins
        self.update_plugins()

        loaded_plugins=list(self.plugin_handler.loaded_plugins.keys())
        for name in loaded_plugins:
            self.plugin_handler.loaded_plugins[name].OnOpenModel(event)

    def OnSimulate(self, event):
        '''OnOpenModel(self, event) --> None
        
        Runs plugin code when the user has simulated the model
        '''
        for name in self.plugin_handler.loaded_plugins:
            self.plugin_handler.loaded_plugins[name].OnSimulate(event)

    def OnFittingUpdate(self, event):
        '''OnOpenModel(self, event) --> None
        
        Runs plugin code when the Fitting algorithm wants to update its output.
        '''
        for name in self.plugin_handler.loaded_plugins:
            self.plugin_handler.loaded_plugins[name].OnFittingUpdate(event)

    def OnGridChanged(self, event):
        """ Runs plugin code to respond to a changed grid.

        :param event:
        :return:
        """
        for name in self.plugin_handler.loaded_plugins:
            self.plugin_handler.loaded_plugins[name].OnGridChange(event)

    def GetPlugin(self, plugin_name):
        return self.plugin_handler.loaded_plugins[plugin_name]

# ==============================================================================
# Utility Dialog functions..




import wx
import wx.adv
import wx.lib.filebrowsebutton as filebrowse
from .help_modules import spec
import numpy as np

from ..data_loader_framework import Template
from ..utils import ShowErrorDialog, ShowWarningDialog
from genx.core.custom_logging import iprint

_maxWidth=450

class Plugin(Template):
    def __init__(self, parent):
        Template.__init__(self, parent)
        self.specfile=None
        self.specfile_name=None
        self.scan=None
        # self.dataset = data.DataSet()
        self.datalist=self.data

    def LoadDataFile(self, selected_items, data_id=0):
        ''' Called when GenX wants to load data'''

        if len(selected_items)==0:
            ShowWarningDialog(self.parent, 'Please select a data set before trying to load a spec file.')
            return False
        old_data=self.datalist[selected_items[0]].copy()
        self.dataset=self.datalist[selected_items[0]]
        wizard=wx.adv.Wizard(self.parent, -1, "Load Spec File",
                             style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER)
        # page1 = TitledPage(wizard, "Page1")
        page1=LoadSpecScanPage(wizard, self,
                               default_filename=self.specfile_name)
        page2=SelectCountersPage(wizard, self)
        page3=CustomManipulationPage(wizard, self)
        page4=SetNamePage(wizard, self)

        wx.adv.WizardPageSimple_Chain(page1, page2)
        wx.adv.WizardPageSimple_Chain(page2, page3)
        wx.adv.WizardPageSimple_Chain(page3, page4)

        wizard.FitToPage(page1)
        # print dir(wizard)
        # print 'data: ', self.datalist[0].x
        if wizard.RunWizard(page1):
            return True
        else:
            self.datalist[selected_items[0]].safe_copy(old_data)
            return False

    def load_specfile(self, filename):
        self.specfile=spec.SpecDataFile(filename)
        names=list(self.specfile.findex.keys())
        names.sort()
        try:
            sc=self.specfile.scan_commands
            names=['%s '%(name,)+sc[name] for name in names]
        except Exception as e:
            iprint("Could not create full names, error: ", e.__str__())
            ShowErrorDialog(self.parent, "Could not create full names, error:\n %s"%e.__str__())
            names=['%s '%name for name in names]
        self.specfile_name=filename
        self.scan_names=names
        return self.scan_names

    def load_scans(self, scanlist):
        '''Function to load the spec scans as a list of scan numbers
        '''
        # self.dataset = data.DataSet()
        try:
            self.scan=self.specfile[scanlist]
            self.dataset.set_extra_data('values', self.scan.values)
        except Exception as e:
            iprint("Could not load the desired scans")
            iprint("Error: ", e.__str__())
            iprint("scanlist: ", scanlist)
            ShowErrorDialog(self.parent, "Could not load the desired scans, error:\n%s "%e.__str__())
        # for key in self.scan.values:
        #    try:
        #        self.dataset.set_extra_data(key, self.scan.values[key])
        #    except Exception, e:
        #        print "Could not load ", key
        #        print error

    def get_data_choices(self):
        '''Function to return the data choices
        '''
        try:
            choices=self.scan.cols
        except Exception as e:
            iprint("Could not load the scan cols error: ", e.__str__())
            ShowErrorDialog(self.parent, "Could not load the scan cols, error:\n%s "%e.__str__())
            choices=[]

        return choices

    def update_data_cols(self, cols, autom=False):
        ''' Update the choices for the different values'''
        iprint(cols)
        self.x_val=cols[0]
        self.det_val=cols[1]
        self.mon_val=cols[2]
        self.error_val=cols[3]
        xval=self.scan.values[self.x_val]
        yvals=[self.scan.values[self.det_val], ]
        if self.mon_val!='None':
            yvals.append(self.scan.values[self.mon_val])
        if self.error_val!='None' and self.error_val!='':
            yvals.append(self.scan.values[self.error_val])
        if autom:
            xval, yvals=automerge(xval, yvals)
        # self.dataset.set_extra_data('det', self.scan.values[self.det_val])
        self.dataset.set_extra_data('det', yvals[0])
        if self.mon_val!='None':
            # self.dataset.set_extra_data('mon', self.scan.values[self.mon_val])
            self.dataset.set_extra_data('mon', yvals[1])

            if self.error_val!='None' and self.error_val!='':
                self.dataset.error_raw=yvals[2]
        elif self.error_val!='None' and self.error_val!='':
            self.dataset.error_raw=yvals[1]

        # self.dataset.x_raw = self.scan.values[self.x_val]
        self.dataset.x_raw=xval
        # if self.error_val != 'None' and self.error_val != '':
        #    self.dataset.error_raw = self.scan.values[self.error_val]

    def get_dataset_names(self):
        return [d.name for d in self.datalist]

    def set_commands(self, commands):
        '''Sets the data commands if valid to commands
        '''
        result=self.dataset.try_commands(commands)
        if result!='':
            ShowErrorDialog(self.parent, result)
            return False
        self.dataset.set_commands(commands)
        self.dataset.run_command()
        return True

class TitledPage(wx.adv.WizardPageSimple):
    def __init__(self, parent, title, plugin):
        wx.adv.WizardPageSimple.__init__(self, parent)
        self.sizer=wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self.sizer)
        titleText=wx.StaticText(self, -1, title)
        titleText.SetFont(wx.Font(18, wx.SWISS, wx.NORMAL, wx.BOLD))
        self.sizer.Add(titleText, 0,
                       wx.ALIGN_CENTRE | wx.ALL, 5)
        self.sizer.Add(wx.StaticLine(self, -1), 0,
                       wx.EXPAND | wx.ALL, 5)

class LoadSpecScanPage(wx.adv.WizardPageSimple):

    def __init__(self, parent, plugin, default_filename=None):
        self.plugin=plugin

        title="Load scan"
        wx.adv.WizardPageSimple.__init__(self, parent)
        self.sizer=wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self.sizer)
        titleText=wx.StaticText(self, -1, title)
        titleText.SetFont(wx.Font(18, wx.SWISS, wx.NORMAL, wx.BOLD))
        self.sizer.Add(titleText, 0,
                       wx.ALIGN_CENTRE | wx.ALL, 5)
        self.sizer.Add(wx.StaticLine(self, -1), 0,
                       wx.EXPAND | wx.ALL, 5)

        self.filebrowser=filebrowse.FileBrowseButton(self,
                                                     -1,
                                                     startDirectory='.',
                                                     size=(450, -1),
                                                     changeCallback=self.filebrowser_callback,
                                                     fileMode=wx.FD_OPEN
                                                     )
        self.filebrowser.textControl.Disable()
        self.sizer.Add(self.filebrowser, 0, wx.EXPAND | wx.ALL, 10)
        chooserText=wx.StaticText(self, -1, "Select the scans to load")
        self.sizer.Add(chooserText, 0,
                       wx.ALIGN_CENTRE | wx.ALL, 5)
        self.scanchooser=wx.CheckListBox(self, -1)
        self.sizer.Add(self.scanchooser, 1, wx.EXPAND | wx.ALL, 2)

        self.Bind(wx.adv.EVT_WIZARD_PAGE_CHANGING, self.OnWizPageChanging)
        self.scanchooser.Bind(wx.EVT_CHECKLISTBOX, self.OnScanSelected)

        self.SetControlEnable(wx.ID_FORWARD, False)

        # Load the defualt spec file -  if exists
        if default_filename:
            self.filebrowser.SetValue(default_filename)
            self.load_file(default_filename)

    def SetControlEnable(self, id, state):
        '''Sets the control enable of window for parent to state. Use wx.ID_FORWARD or
        wx.ID_BACKWARD for the beck forward buttons.
        '''
        win=self.Parent.FindWindowById(id)
        if win:
            win.Enable(state)
        else:
            raise Exception('Can not find window %s'%(id,))

    def OnScanSelected(self, evt):
        ''' A scan has been seelcted'''
        # check so at least one scan is selected
        if len(self.scanchooser.GetChecked())>0:
            self.SetControlEnable(wx.ID_FORWARD, True)
        else:
            self.SetControlEnable(wx.ID_FORWARD, False)

    def OnWizPageChanging(self, evt):
        # Check if we are moving forward
        if evt.GetDirection():
            scan_numbers=self.extract_scan_numbers()
            if self.plugin.specfile:
                try:
                    self.plugin.load_scans(scan_numbers)
                except Exception as e:
                    ShowErrorDialog(self.Parent, 'Could not load the selected scan'
                                                 ' - The scan might be corrupt.\n'
                                                 'Please choose another.\n'
                                                 'Python error: '+e.__str__())
                    # print e.__str__()
                    evt.Veto()

    def extract_scan_numbers(self):
        '''Extract the scan numbers from the Checklistbox
        '''
        scan_numbers=self.scanchooser.CheckedStrings
        # Extract the numbers from the strings
        scan_numbers=[int(ss.split(' ')[0]) for ss in scan_numbers]
        return scan_numbers

    def load_file(self, filename):
        self.scanchooser.Set(self.plugin.load_specfile(filename))
        # self.SetControlEnable(wx.ID_FORWARD, True)
        # print filename

    def filebrowser_callback(self, evt):
        filename=evt.GetString()
        self.load_file(filename)

class SelectCountersPage(wx.adv.WizardPageSimple):
    def __init__(self, parent, plugin):
        wx.adv.WizardPageSimple.__init__(self, parent)
        self.plugin=plugin
        self.sizer=wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self.sizer)
        title="Select Counters/Motors"
        titleText=wx.StaticText(self, -1, title)
        titleText.SetFont(wx.Font(18, wx.SWISS, wx.NORMAL, wx.BOLD))
        self.sizer.Add(titleText, 0,
                       wx.ALIGN_CENTRE | wx.ALL, 5)
        self.sizer.Add(wx.StaticLine(self, -1), 0,
                       wx.EXPAND | wx.ALL, 5)

        self.choiceSizer=wx.FlexGridSizer(rows=4, cols=2,
                                          vgap=10, hgap=10)
        self.choiceSizer.Add(wx.StaticText(self, -1, "X value"),
                             0, wx.EXPAND | wx.ALL, 5)
        self.xChoice=wx.Choice(self, -1, (-1, -1), choices=[])
        self.choiceSizer.Add(self.xChoice, 0, wx.EXPAND | wx.ALL, 5)
        self.choiceSizer.Add(wx.StaticText(self, -1, "Det. value"),
                             0, wx.EXPAND | wx.ALL, 5)
        self.detChoice=wx.Choice(self, -1, (-1, -1), choices=[])
        self.choiceSizer.Add(self.detChoice, 0, wx.EXPAND | wx.ALL, 5)
        self.choiceSizer.Add(wx.StaticText(self, -1, "Mon. value"),
                             0, wx.EXPAND | wx.ALL, 5)
        self.monChoice=wx.Choice(self, -1, (-1, -1), choices=[])
        self.choiceSizer.Add(self.monChoice, 0, wx.EXPAND | wx.ALL, 5)
        self.choiceSizer.Add(wx.StaticText(self, -1, "Error value"),
                             0, wx.EXPAND | wx.ALL, 5)
        self.errorChoice=wx.Choice(self, -1, (-1, -1), choices=[])
        self.choiceSizer.Add(self.errorChoice, 0, wx.EXPAND | wx.ALL, 5)
        self.sizer.Add(self.choiceSizer, 0, wx.ALIGN_CENTRE | wx.ALL)
        self.automergeCheckBox=wx.CheckBox(self, -1, "Auto merge data")
        self.automergeSizer=wx.FlexGridSizer(rows=1, cols=1,
                                             vgap=10, hgap=10)
        self.automergeSizer.Add(self.automergeCheckBox, 0, wx.EXPAND | wx.ALL, 5)
        self.sizer.Add(self.automergeSizer, 0, wx.ALIGN_CENTRE | wx.ALL)

        self.Bind(wx.adv.EVT_WIZARD_PAGE_CHANGED, self.OnWizPageChanged)
        self.Bind(wx.adv.EVT_WIZARD_PAGE_CHANGING, self.OnWizPageChanging)

    def OnWizPageChanged(self, evt):
        # Check if we are moving forward
        # print 'In wiz page changing ', evt.GetDirection()
        if evt.GetDirection():
            choices=self.plugin.get_data_choices()
            self.update_counters(choices)

    def OnWizPageChanging(self, evt):
        if evt.GetDirection():
            self.plugin.update_data_cols(self.get_selected(),
                                         self.automergeCheckBox.IsChecked())

    def update_counters(self, mot_count):
        self.xChoice.SetItems(mot_count)
        if len(mot_count)>0:
            self.xChoice.SetSelection(0)
        self.detChoice.SetItems(mot_count)
        try:
            self.detChoice.SetSelection(mot_count.index('Detector'))
        except:
            self.detChoice.SetSelection(0)
        self.monChoice.SetItems(['None']+mot_count)
        try:
            self.monChoice.SetSelection(mot_count.index('Monitor')+1)
        except:
            self.monChoice.SetSelection(0)
        self.errorChoice.SetItems(['None']+mot_count)
        self.errorChoice.SetSelection(0)

    def get_selected(self):
        data=(self.xChoice.GetStringSelection(),
              self.detChoice.GetStringSelection(),
              self.monChoice.GetStringSelection(),
              self.errorChoice.GetStringSelection(),
              )
        return data

class CustomManipulationPage(wx.adv.WizardPageSimple):
    def __init__(self, parent, plugin, data_sets=None):
        wx.adv.WizardPageSimple.__init__(self, parent)
        if data_sets is None:
            data_sets=[]
        self.plugin=plugin
        self.sizer=wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self.sizer)
        title="Custom Manipulation"
        titleText=wx.StaticText(self, -1, title)
        titleText.SetFont(wx.Font(18, wx.SWISS, wx.NORMAL, wx.BOLD))
        self.sizer.Add(titleText, 0,
                       wx.ALIGN_CENTRE | wx.ALL, 5)
        self.sizer.Add(wx.StaticLine(self, -1), 0,
                       wx.EXPAND | wx.ALL, 5)
        help_text=("By changing the expressions in the boxes below you can "+
                   "manipulate your data with expressions written in Python.\n"+
                   "These transforms can be changed later from the data calculations"+
                   "dialog.")
        helpText=wx.StaticText(self, -1, help_text)
        helpText.Wrap(_maxWidth)
        self.sizer.Add(helpText, 0, wx.ALIGN_CENTRE | wx.ALL, 10)
        self.manipSizer=wx.FlexGridSizer(rows=4, cols=2,
                                         vgap=10, hgap=10)
        self.manipSizer.Add(wx.StaticText(self, -1, "x = "),
                            0, wx.EXPAND | wx.ALL, 5)
        self.xCtrl=wx.TextCtrl(self, -1, value='x')
        self.manipSizer.Add(self.xCtrl, 0, wx.EXPAND | wx.ALL, 5)

        self.manipSizer.Add(wx.StaticText(self, -1, "y = "),
                            0, wx.EXPAND | wx.ALL, 5)
        self.yCtrl=wx.TextCtrl(self, -1, value='det')
        self.manipSizer.Add(self.yCtrl, 0, wx.EXPAND | wx.ALL, 5)

        self.manipSizer.Add(wx.StaticText(self, -1, "Error = "),
                            0, wx.EXPAND | wx.ALL, 5)
        self.errorCtrl=wx.TextCtrl(self, -1, value='0*x')
        self.manipSizer.Add(self.errorCtrl, 0, wx.EXPAND | wx.ALL, 5)
        self.sizer.Add(self.manipSizer, 0, wx.ALIGN_CENTRE | wx.ALL)

        help_text="Import the settings from data set: "
        helpText=wx.StaticText(self, -1, help_text)
        helpText.Wrap(_maxWidth)
        self.sizer.Add(helpText, 0, wx.ALIGN_CENTRE | wx.ALL, 10)
        self.importChoice=wx.Choice(self, -1, (-1, 50), choices=data_sets)
        self.sizer.Add(self.importChoice, 0, wx.ALIGN_CENTRE | wx.ALL, 10)
        self.Bind(wx.EVT_CHOICE, self.EvtImport, self.importChoice)

        self.Bind(wx.adv.EVT_WIZARD_PAGE_CHANGED, self.OnWizPageChanged)
        self.Bind(wx.adv.EVT_WIZARD_PAGE_CHANGING, self.OnWizPageChanging)

    def OnWizPageChanged(self, evt):
        ''' Event handler for entering this page'''
        manip_string=['x', 'det']
        if self.plugin.mon_val!='None':
            manip_string.append('mon')
        else:
            manip_string.append('None')
        if self.plugin.error_val!='None':
            manip_string.append('e')
        else:
            manip_string.append('None')

        self.set_manip(manip_string)
        self.importChoice.SetItems(self.plugin.get_dataset_names())

    def OnWizPageChanging(self, evt):
        ''' Event Handler for leaving this page'''
        commands={'x': self.xCtrl.GetValue(),
                  'y': self.yCtrl.GetValue(),
                  'e': self.errorCtrl.GetValue(), }
        if not self.plugin.set_commands(commands):
            evt.Veto()

    def EvtImport(self, evt):
        ''' On the event to import manipulations from another data set'''
        item=self.importChoice.GetSelection()
        com=self.plugin.datalist[item].get_commands()
        # self.set_manip((com['x'], com['y'], com['e']))
        self.xCtrl.ChangeValue(com['x'])
        self.yCtrl.ChangeValue(com['y'])
        self.errorCtrl.ChangeValue(com['e'])

    def set_manip(self, choices):
        '''Sets the manipulations to defualt values as given by
        the choices in the form (x, det, mon, error)
        '''
        iprint(choices)
        self.xCtrl.ChangeValue(choices[0])
        if choices[2]=='None':
            self.yCtrl.ChangeValue(choices[1])
            if choices[3]=='None':
                self.errorCtrl.ChangeValue("sqrt(%s)"
                                           %(choices[1],))
            else:
                self.errorCtrl.ChangeValue(choices[3])
        else:
            self.yCtrl.ChangeValue("%s/%s"%(choices[1], choices[2]))
            if choices[3]=='None':
                self.errorCtrl.ChangeValue("sqrt(1.0/%s + 1.0/%s)*%s/%s"
                                           %(choices[1], choices[2],
                                             choices[1], choices[2]))
            else:
                self.errorCtrl.ChangeValue(choices[3])

class SetNamePage(wx.adv.WizardPageSimple):

    def __init__(self, parent, plugin, data_sets=None):
        wx.adv.WizardPageSimple.__init__(self, parent)
        if data_sets is None:
            data_sets=[]
        self.plugin=plugin
        self.sizer=wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self.sizer)
        title="Name"
        titleText=wx.StaticText(self, -1, title)
        titleText.SetFont(wx.Font(18, wx.SWISS, wx.NORMAL, wx.BOLD))
        self.sizer.Add(titleText, 0,
                       wx.ALIGN_CENTRE | wx.ALL, 5)
        self.sizer.Add(wx.StaticLine(self, -1), 0,
                       wx.EXPAND | wx.ALL, 5)
        help_text=("Set the name of the data set as it will appear in GenX."
                   "The name can be changed later by a slow double click on the name "
                   "in the data panel.")
        helpText=wx.StaticText(self, -1, help_text)
        helpText.Wrap(_maxWidth)
        self.sizer.Add(helpText, 0, wx.ALIGN_CENTRE | wx.ALL, 10)

        self.manipSizer=wx.FlexGridSizer(rows=1, cols=2,
                                         vgap=10, hgap=10)
        self.manipSizer.Add(wx.StaticText(self, -1, "Name: "),
                            0, wx.EXPAND | wx.ALIGN_CENTER_VERTICAL | wx.ALL, 10)
        self.xCtrl=wx.TextCtrl(self, -1, value='Name')
        self.manipSizer.Add(self.xCtrl, 0, wx.EXPAND | wx.ALL, 10)
        self.sizer.Add(self.manipSizer, 0, wx.EXPAND | wx.ALIGN_CENTRE | wx.ALL)

        self.Bind(wx.adv.EVT_WIZARD_PAGE_CHANGING, self.OnWizPageChanging)

    def OnWizPageChanging(self, evt):
        ''' Event Handler for leaving this page'''

        self.plugin.dataset.name=self.xCtrl.GetValue()


def automerge(xval, yvals, rel_cond=100.0):
    ''' Merges the values according to rel_cond where the 
    the merging criterion is the distance between the two first datapoints divided by
    rel_cond.  
    
    xval - an arraray of x values
    y-vals a tuple of yvals
    rel_cond as above
    '''
    min_size=np.abs(xval[1]-xval[0])/rel_cond
    xarg=xval.argsort()
    xnew=np.array([xval[xarg[0]]])*1.0
    ynew=[]
    for yval in yvals:
        ynew.append(np.array([yval[xarg[0]]])*1.0)
    inew=0
    for i in range(len(xarg))[1:]:
        if np.abs(xnew[-1]-xval[xarg[i]])<=min_size:
            for j in range(len(ynew)):
                ynew[j][-1]+=yvals[j][xarg[i]]
        else:
            xnew=np.append(xnew, xval[xarg[i]])
            for j in range(len(ynew)):
                ynew[j]=np.append(ynew[j], yvals[j][xarg[i]])
            inew+=1

    return xnew, ynew

if __name__=="__main__":
    app=wx.SimpleApp()
    plugin=Plugin(-1)
    wizard=wx.adv.Wizard(None, -1, "Simple Wizard",
                         style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER)
    # page1 = TitledPage(wizard, "Page1")
    page2=LoadSpecScanPage(wizard, plugin)
    page3=SelectCountersPage(wizard, plugin)
    page4=CustomManipulationPage(wizard, plugin)
    page5=SetNamePage(wizard, plugin)

    # page1.sizer.Add(wx.StaticText(page1, -1, "Testing the wizard"))
    # page4.sizer.Add(wx.StaticText(page4, -1, "This is the last page"))

    # wx.adv.WizardPageSimple_Chain(page1, page2)
    wx.adv.WizardPageSimple_Chain(page2, page3)
    wx.adv.WizardPageSimple_Chain(page3, page4)
    wx.adv.WizardPageSimple_Chain(page4, page5)

    wizard.FitToPage(page2)
    # print dir(wizard)
    if wizard.RunWizard(page2):
        iprint("Sucess")



'''<h1> Test </h1>
A plugin to test the builtin features for the plugin.
Opens three new tabs, one in each area of the main window.
It also adds a a custom menu.
'''
from .. import add_on_framework as framework
from genx.core.custom_logging import iprint

class Plugin(framework.Template):
    def __init__(self, parent):
        framework.Template.__init__(self, parent)
        plotpanel=self.NewPlotFolder('Test')
        inputpanel=self.NewInputFolder('Test')
        datapanel=self.NewDataFolder('Test')
        menu=self.NewMenu('Test')
        iprint('Everyting tested, should be visible :-)')



# -*- coding: utf8 -*-
'''
Library for combined x-ray and neutrons simulations for inhomogeneous samples.
==============================================================================
In addition to the options from spec_nx this allows to model thickness variation over
the sample surface and gradients in the roughness and thickness over a repetition of layers.

Classes
-------

Layer
~~~~~
``Layer(b = 0.0, d = 0.0, f = 0.0+0.0J, dens = 1.0, magn_ang = 0.0, magn = 0.0, sigma = 0.0, sigma_gradient=0.0, d_gradient=0.0)``

``b``
   The neutron scattering length per formula unit in fm (fermi meter =
   1e-15m)
``d``
   The thickness of the layer in AA (Angstroms = 1e-10m)
``f``
   The x-ray scattering length per formula unit in electrons. To be
   strict it is the number of Thompson scattering lengths for each
   formula unit.
``dens``
   The density of formula units in units per Angstroms. Note the units!
``magn_ang``
   The angle of the magnetic moment in degress. 0 degrees correspond to
   a moment collinear with the neutron spin.
``magn``
   The magnetic moment per formula unit (same formula unit as b and dens
   refer to)
``sigma``
   The root mean square roughness of the top interface of the layer in
   Angstroms.
``xs_ai``
   The sum of the absorption cross section and the incoherent scattering
   cross section in barns for neutrons
``sigma_gradient``
   Increase of roughness of this layer from bottom to top in stack repetitions
``d_gradient``
   Increase of the thickness of this layer from bottom to top in stack repetitions

Stack
~~~~~
``Stack(Layers = [], Repetitions = 1, sigma_gradient=0.0, sigma_gtype=3, d_gradient=0.0, dens_gradient=0.0)``

``Layers``
   A ``list`` consiting of ``Layer``\ s in the stack the first item is
   the layer closest to the bottom
``Repetitions``
   The number of repsetions of the stack
``sigma_gradient``
   amount of increase in roughness from bottom to top applied to all layers in the stack
``sigma_gtype``
   model for this increase in roughness (0: relative linlinea, 1: relative sqrt, 2:absolute linear, 3: absolute sqrt)
``d_gradient``
   thickness increase from bottom to top
``beta_sm``
   Parameter used to model neutron super-mirror coatings. There are two models implemented.
   If *beta_sm* is positive, a simplified analystical model from J. Schelten and K. Mika (Nuc. Inst. Metho. 160 (1979))
   is used and the parameter represents the quality paramter for super-mirror sequences.
   A super-mirror sequence is generated and Repetitions is interpreted as the m-value for
   the sequence to calculate actual repetitions automatically.

   If *beta_sm* is negative it is interpretated as the zeta parameter of an iterative build-up of super-mirror
   layers using the method described by J.B. Hayter and H.A. Mook (J. Appl. Cryst. (1989), 22, 35-41). Here,
   the actual given number of repetitions are used to build up a super-mirror from the critical edge
   upowards. The m-value follows from the number of repetitions, material parameters and zeta.

   This method only works acuratly with two layers in the Stack, as the maximum and
   minimum SLD layers are used for the SLD calculation needed to get the layer sequences.
   (Other layers are simulated but their SLDs are ignored for the
   number of repetitions and super-lattice repetition period).

   For the Schelten/Mika method, all layers are scaled to the relative size needed to produce the
   right super-lattice period. In Hayter/Mook only the two min/max components are set to the
   optimal thickness values while the rest of the stack stays with constant width.

   The value of d_gradient is interpreted as a relative change from bottom to top in this case. Together
   with the *sm_scale*, that is applied to all layers.
``sm_scale``
   A scaling parameter applied to all thicknesses in a supermirror. This can account for thickness
   differences introduced by the manufacturing process due to imperfect calibration.
``dens_gradient``
   density increase from bottom to top

Sample
~~~~~~
``Sample(Stacks = [], Ambient = Layer(), Substrate = Layer(), sigma_inhom=0.0, lscale_inhom=0.9, flatwidth_inhom=0.3, steps_inhom=20, type_inhom='semi-gauss')``

``Stacks``
   A ``list`` consiting of ``Stack``\ s in the stacks the first item is
   the layer closest to the bottom
``Ambient``
   A ``Layer`` describing the Ambient (enviroment above the sample).
   Only the scattering lengths and density of the layer is used.
``Substrate``
   A ``Layer`` describing the substrate (enviroment below the sample).
   Only the scattering lengths, density and roughness of the layer is
   used.
``sigma_inhom``
   Width of the thickness distribution
``lscale_inhom``
   For empirical PLD model this defines the width of the increas peak
``flatwidth_inhom``
   For empirical PLD model this defines the trnsition from flat to increased probability.
``steps_inhom``
   Number of simulations to be performed to sample the thickness distribution model
``type_inhom``
   Function for the thickness probability. Either symmetric gaussian, semi-gaussian which does not have any probability
   for a higher thickness and an empirical model developed from PLD samples.
``crop_sld``
   Useful for multilayers with a very large number of repetitions. Only keeps crop_sld number of layers
   on top and bottom and removes all in the center for the SLD plot. The removed layers are replace with
   one layer of empty space. If the parameter is negative this gap can be replaced by a "peak" that
   separates top and bottom, this does invalidate the x-axis of the SLD plot, though.

The empricial PLD model is base of simulations from x-ray line focus plume shapes to get thickness
distribution probabilities and reduced to a smaller number of parameters.
See PhD thesis *Multiferroicity in oxide thin films and heterostructures, A.Glavic, RWTH Aachen, (2012)* for a
more detailed description of the empirical PLD model and an example application.

Instrument
~~~~~~~~~~
``Instrument(probe = 'x-ray', wavelength = 1.54, coords = '2θ', I0 = 1.0 res = 0.001, restype = 'no conv', respoints = 5, resintrange = 2, beamw = 0.01, footype = 'no corr', samplelen = 10.0, incangle = 0.0, pol = 'uu')``

``probe``
    Describes the radiation and measurments used is one of: 'x-ray',
    'neutron', 'neutron pol', 'neutron pol spin flip', 'neutron tof',
    'neutron pol tof' or the respective number 0, 1, 2, 3, 4, 5, 6. The
    calculations for x-rays uses ``f`` for the scattering length for
    neutrons ``b`` for 'neutron pol', 'neutron pol spin flip' and 'neutron
    pol tof' alternatives the ``magn`` is used in the calculations. Note
    that the angle of magnetization ``magn_ang`` is only used in the last
    alternative.
``wavelength``
    The wavelength of the radiation given in AA (Angstroms)
``coords``
    The coordinates of the data given to the SimSpecular function. The
    available alternatives are: 'q' or '2θ'. Alternatively the numbers 0 (q)
    or 1 (tth) can be used.
``I0``
    The incident intensity (a scaling factor)
``Ibkg``
    The background intensity. Added as a constant value to the calculated
    reflectivity
``res``
    The resolution of the instrument given in the coordinates of ``coords``.
    This assumes a gaussian resolution function and ``res`` is the standard
    deviation of that gaussian. If ``restype`` has (dx/x) in its name the
    gaussian standard deviation is given by res*x where x is either in tth
    or q.
``restype``
    Describes the rype of the resolution calculated. One of the
    alterantives: 'no conv', 'fast conv', 'full conv and varying res.',
    'fast conv + varying res.', 'full conv and varying res. (dx/x)', 'fast
    conv + varying res. (dx/x)'. The respective numbers 0-3 also works. Note
    that fast convolution only alllows a single value into res wheras the
    other can also take an array with the same length as the x-data (varying
    resolution)
``respoints``
    The number of points to include in the resolution calculation. This is
    only used for 'full conv and vaying res.', 'fast conv + varying res',
    'full conv and varying res. (dx/x)' and 'fast conv + varying res.
    (dx/x)'.
``resintrange``
    Number of standard deviatons to integrate the resolution function times
    the reflectivity over
``footype``
    Which type of footprint correction is to be applied to the simulation.
    One of: 'no corr', 'gauss beam' or 'square beam'. Alternatively, the
    number 0-2 are also valid. The different choices are self expnalatory.
``beamw``
    The width of the beam given in mm. For 'gauss beam' it should be the
    standard deviation. For 'square beam' it is the full width of the beam.
``samplelen``
    The length of the sample given in mm
``incangle``
    The incident angle of the neutrons, only valid in tof mode
``pol``
    The measured polarization of the instrument. Valid options are:
    'uu','dd', 'ud', 'du' or 'ass' the respective number 0-3 also works.
'''
from copy import deepcopy
from numpy import *
from scipy.special import wofz
from . import spec_nx
from .lib.sm_hayter_mook import sm_layers
from .spec_nx import refl
from .lib.instrument import *

# Preamble to define the parameters needed for the models outlined below:
ModelID='SpecInhom'
__pars__=spec_nx.__pars__.copy()

instrument_string_choices=spec_nx.instrument_string_choices.copy()
InstrumentParameters=spec_nx.InstrumentParameters.copy()
InstrumentGroups=spec_nx.InstrumentGroups.copy()
InstrumentUnits=spec_nx.InstrumentUnits.copy()

LayerParameters=spec_nx.LayerParameters.copy()
LayerParameters.update({'sigma_gradient': 0.0, 'd_gradient': 0.0})
LayerUnits=spec_nx.LayerUnits.copy()
LayerUnits.update({'sigma_gradient': '1', 'd_gradient': '1'})
LayerGroups=[('Standard', ['f', 'dens', 'd', 'sigma']),
             ('Neutron', ['b', 'xs_ai', 'magn', 'magn_ang']),
             ('Inhom.', ['sigma_gradient', 'd_gradient'])]

StackParameters={'Layers': [], 'Repetitions': 1,
                 'sigma_gradient': 0.,  # amount of increase in roughness from bottom to top
                 'sigma_gtype': 3,  # model for this increas (rel lin, rel sqrt, abs lin, abs sqrt)
                 'd_gradient': 0.,  # change in thickness bottom to top
                 'beta_sm' : 0.0, # function applied to change thickness (linear,  super-mirror (d=beta SM))
                 'sm_scale': 1.0, # scaling factor for thicknesses in super-mirror
                 'dens_gradient': 0.,  # change in density bottom to top
                 }
SampleParameters=spec_nx.SampleParameters.copy()
SampleParameters.update({'sigma_inhom': 0.0, 'lscale_inhom': 0.9, 'flatwidth_inhom': 0.3,
                         'steps_inhom': 20, 'type_inhom': 'empiric PLD', 'crop_sld': 200})
sample_string_choices={
    'type_inhom': ['gauss', 'semi-gauss', 'empiric PLD'],
    }

AA_to_eV=spec_nx.AA_to_eV
q_limit=spec_nx.q_limit
Buffer=spec_nx.Buffer

__xlabel__ = "q [Å$^{-1}$]"
__ylabel__ = "Instnsity [a.u.]"

def Specular(TwoThetaQz, sample, instrument):
    '''
      The model function. Averadging the intensities for different
      layer thicknesses as found for e.g. large PLD samples.
    '''
    global __xlabel__

    Q, TwoThetaQz, weight=spec_nx.resolution_init(TwoThetaQz, instrument)
    if any(Q<q_limit):
        raise ValueError('The q vector has to be above %.1e'%q_limit)
    restype=instrument.getRestype()
    foottype=instrument.getFootype()
    Ibkg=instrument.getIbkg()
    instrument.setRestype('no conv')
    instrument.setFootype('no corr')
    instrument.Ibkg=0.
    sampcall=deepcopy(sample)

    # average thicknesses before inhomogeniety average
    d0=[array([Layer.getD() for Layer in Stack.Layers]) for Stack in sample.Stacks]
    sigma_d=sample.getSigma_inhom()*0.01  # Inhomogeniety in \% (gamma for type 2)
    lorentz_scale=sample.getLscale_inhom()
    flat_width=maximum(1e-4, sample.getFlatwidth_inhom()*0.01)
    type_inhom=sample.getType_inhom()
    # Define the thicknesses to calculate and their propability
    if type_inhom in sample_string_choices['type_inhom']:
        type_inhom=sample_string_choices['type_inhom'].index(type_inhom)
    else:
        type_inhom=0
    if sigma_d==0:  # no inhomogeniety
        d_fact=array([1.])
        P=array([1.])
    elif type_inhom==1:  # half gaussian shape inhomogeniety
        d_fact=1.+linspace(-2.*sigma_d, 0, sample.getSteps_inhom())
        P=exp(-0.5*(d_fact-sigma_d-1.)**2/sigma_d**2)
        P/=P.sum()
        mean_d=(P*d_fact).sum()
        d_fact+=1.-mean_d
    elif type_inhom==2:  # inhomogeniety of a PLD line focus, better approximation
        d_fact=1.+linspace(-1.*max(2.*sigma_d, flat_width), 0, sample.getSteps_inhom())
        Pg=where(d_fact>flat_width, lorentz_scale*1./(1.+((d_fact-1.)/sigma_d)**2), 0.)
        Pf=(1.-lorentz_scale)*where(d_fact>flat_width, 1., 0.)
        P=Pg+Pf
        P/=P.sum()
        mean_d=(P*d_fact).sum()
        d_fact+=1.-mean_d
    else:  # gaussian inhomegeniety
        d_fact=1.+linspace(-sigma_d, sigma_d, sample.getSteps_inhom())
        P=exp(-0.5*(d_fact-1.)**2/sigma_d**2)
        P/=P.sum()
    # list for reflectivities to average
    Rlist=[]
    # Iterate over thicknesses
    for d_facti, Pi in zip(d_fact, P):
        di=[d_facti*d0i for d0i in d0]
        for i, Stack in enumerate(sampcall.Stacks):
            for j, Layer in enumerate(Stack.Layers):
                # Layer.setD(di[i][j])
                Layer.d=di[i][j]
        Rlist.append(Pi*spec_nx.Specular(TwoThetaQz, sampcall, instrument))
    R=array(Rlist).sum(axis=0)

    instrument.setRestype(restype)
    instrument.setFootype(foottype)
    instrument.Ibkg=Ibkg

    # footprint correction
    foocor=spec_nx.footprintcorr(Q, instrument)
    # resolution correction
    R=spec_nx.resolutioncorr(R, TwoThetaQz, foocor, instrument, weight)

    __xlabel__ = spec_nx.__xlabel__
    return R+instrument.getIbkg()

def ResolutionVectorAsymetric(Q, dQ, points, dLambda, asymmetry, range_=3):
    '''
      Resolution vector for a asymmetric wavelength distribution found in
      neutron experiments with multilayer monochromator.
    '''
    Qrange=max(range_*dQ, range_*dLambda*Q.max())
    Qstep=2*Qrange/points
    Qres=Q+(arange(points)-(points-1)/2)[:, newaxis]*Qstep
    Quse=transpose(Q[:, newaxis])

    gamma_asym=2.*dLambda*Quse/(1+exp(asymmetry*(Quse-Qres)))
    z=(Quse-Qres+(abs(gamma_asym)*1j))/abs(dQ)/sqrt(2.)
    z0=(0.+(abs(gamma_asym)*1j))/abs(dQ)/sqrt(2)
    weight=wofz(z).real/wofz(z0).real
    Qret=Qres.flatten()
    return Qret, weight

POL_CHANNELS = ['uu', 'ud', 'du', 'dd']

def PolSpecular(TwoThetaQz, p1, p2, F1, F2, sample, instrument):
    """
    Specular reflectivity of polarized measurement with finite polarization.
    The polarization parameters are in accordance to the definition used in
    A.R. Wildes publication, Review of Scientific Instruments 70, 11 (1999)
    https://doi.org/10.1063/1.1150060

    pol_params = (p1, p2, f1, f2)
    p1: polarizer efficiency, 0: 100% spin-up, 1: 100% spin-down, 0.5: unpolarized
    p2: analyzer efficiency,  0: 100% spin-up, 1: 100% spin-down, 0.5: unpolairzed
    F1/F2: Flipper efficienty, 0: 100% efficient, 1: no flipping

    # BEGIN Parameters
    TwoThetaQz data.x
    p1 0.
    p2 0.
    F1 0.
    F2 0.
    # END Parameters

    """
    inst_pol = instrument.pol
    if not inst_pol in POL_CHANNELS:
        raise ValueError(f"Instrument polarization as to be one of {POL_CHANNELS}.")
    if instrument.probe not in [3, instrument_string_choices['probe'][3]]:
        raise ValueError("Polarization corrected simulation requires probe to be 'neutron pol spin flip'")

    instrument.pol = 'uu'
    uu = Specular(TwoThetaQz, sample, instrument)
    instrument.pol = 'dd'
    dd = Specular(TwoThetaQz, sample, instrument)
    instrument.pol = 'ud'
    ud = Specular(TwoThetaQz, sample, instrument)
    du = ud
    instrument.pol = inst_pol

    P = get_pol_matrix(p1, p2, F1, F2)
    Pline = P[POL_CHANNELS.index(instrument.pol)]
    I = Pline[:, newaxis] * vstack([uu, ud, du, dd])
    return I.sum(axis=0)

SLD_calculations=spec_nx.SLD_calculations

SimulationFunctions={'Specular': Specular,
                     'PolSpecular': PolSpecular,
                     'SLD': SLD_calculations
                     }

(Instrument, Layer, Stack, Sample)=refl.MakeClasses(InstrumentParameters,
                                                    LayerParameters, StackParameters, SampleParameters,
                                                    SimulationFunctions,
                                                    ModelID)

# Add gradient for sigma and thickness to multilayers
def resolveLayerParameter(self, parameter):
    if self.beta_sm>0:
        # user has selected super-mirror coating sequence
        m = self.Repetitions
        beta = self.beta_sm
        rho_all = [refl.resolve_par(L, 'dens')*(refl.resolve_par(L, 'b').real+refl.resolve_par(L, 'magn'))*1e-5
                   for L in self.Layers]
        rhoA = max(rho_all)*1e-5
        rhoB = min(rho_all)*1e-5
        # critical wave vector values (in paper this is includes an addition factor of 1/2)
        QcA = 4.*pi*sqrt(rhoA/pi+0j)
        QcB = 4.*pi*sqrt(rhoB/pi+0j)
        Qcm = 0.022*m # critical q for given m-value
        g=Qcm/QcA
        QAB = QcA / sqrt(QcA**2-QcB**2)
        b = beta*QAB**4 + 2.*beta*QAB**2 - 1
        N = int(abs(beta*((g**2*QAB**2+1)**2-1)-b))
    elif self.beta_sm<0:
        # super-mirror from Hayter and Mook iterative method was selected, fixed number of layers
        N = int(abs(self.Repetitions))
    else:
        N = int(self.Repetitions)
    if parameter=='sigma':
        sigma_gradient=self.sigma_gradient
        # parameters for layers with roughness gradient
        par=[refl.resolve_par(lay, parameter) for lay in self.Layers]
        for i in range(1, N):
            if self.sigma_gtype==0:
                # linear increase of roughness to (1+sigma_gradient) times bottom roughness)
                par+=[refl.resolve_par(lay, parameter)*(1.+(sigma_gradient+lay.sigma_gradient)*i/(N-1))
                      for lay in self.Layers]
            elif self.sigma_gtype==1:
                # add roughness using rms
                par+=[refl.resolve_par(lay, parameter)*sqrt(
                    1.+((sigma_gradient+lay.sigma_gradient)*i/(N-1))**2)
                      for lay in self.Layers]
            elif self.sigma_gtype==2:
                # linear increase of roughness to bottom roughness + sigma_gradient)
                par+=[refl.resolve_par(lay, parameter)+(sigma_gradient+lay.sigma_gradient)*i/(N-1)
                      for lay in self.Layers]
            elif self.sigma_gtype==3:
                # add roughness using rms
                par+=[sqrt(
                    refl.resolve_par(lay, parameter)**2+((sigma_gradient+lay.sigma_gradient)*i/(N-1))**2)
                      for lay in self.Layers]
            else:
                raise NotImplementedError('sigma_gtype must be between 0 and 3')
    elif parameter=='d':
        d_gradient = self.d_gradient
        if self.beta_sm>0:
            # layer thickness sequence based on calculated bi-layer thicknesses for super-mirror
            L_thicknesses = [refl.resolve_par(lay, 'd')+0.0 for lay in self.Layers]
            D_start = sum(L_thicknesses)
            rel_thickness = [di/D_start for di in L_thicknesses]
            sm_scale=self.sm_scale
            idx=arange(N)
            D_SL=abs((2.*QAB*pi/QcA) / sqrt(sqrt(1+(N-idx+b)/beta)-1))*(1.0+d_gradient*idx/(N-1))*sm_scale
            par = []
            for i in range(N):
                par += [di*D_SL[i] for di in rel_thickness]
        elif self.beta_sm<0:
            zeta = -self.beta_sm
            rho_all = [refl.resolve_par(L, 'dens')*(refl.resolve_par(L, 'b').real+refl.resolve_par(L, 'magn'))*1e-5
                       for L in self.Layers]
            rho1 = max(rho_all) # in paper, SLD is called alpha
            rho2 = min(rho_all)
            # if two layers are chosen, use scale from olgorithm, else only change main 2 layers size
            L_thicknesses = [refl.resolve_par(lay, 'd')+0.0 for lay in self.Layers]
            D_start = sum(L_thicknesses)
            idx1,idx2=rho_all.index(rho1),rho_all.index(rho2)
            rest_size=D_start-L_thicknesses[idx1]-L_thicknesses[idx2]
            # layer thicknesses according to Hayter+Mook algorithm
            D_SL=sm_layers(rho1, rho2, N, zeta)
            # sign on repetitions defines direction of multilayer
            if self.Repetitions>0:
                D_SL.reverse()
            par=[]
            sm_scale=self.sm_scale
            for i, (d1, d2) in enumerate(D_SL):
                # variation from user parameters
                var_scale=(1.0+d_gradient*i/(N-1))*sm_scale
                # scale the size of main two layer but remove residual layer thicknesses
                total_bilayer=d1+d2
                scaled_bilayer=(total_bilayer-rest_size)/total_bilayer
                for i, di in enumerate(L_thicknesses):
                    if i==idx1:
                        par.append(d1*scaled_bilayer*var_scale)
                    elif i==idx2:
                        par.append(d2*scaled_bilayer*var_scale)
                    else:
                        par.append(di*var_scale)
        else:
            # parameters for layers with thickness gradient
            par=[]
            for i in range(N):
                par+=[refl.resolve_par(lay, parameter)*(1.-(d_gradient+lay.d_gradient)*(1./2.-float(i)/N))
                      for lay in self.Layers]
    elif parameter in ['dens']:
        dens_gradient=self.dens_gradient
        # parameters for layers with roughness gradient
        par=[]
        for i in range(N):
            par+=[refl.resolve_par(lay, parameter)*(1.-dens_gradient*(1./2.-float(i)/self.Repetitions)) for lay in
                  self.Layers]
    else:
        par=[refl.resolve_par(lay, parameter)+0.0 for lay in self.Layers]*N
    return par

Stack.resolveLayerParameter=resolveLayerParameter



"""
GenX main package.
"""
# make sure the models can import subpackages correctly
import os
import sys
from importlib import import_module
from importlib.abc import MetaPathFinder
from importlib.util import spec_from_file_location

_buildin_metafinder = list(sys.meta_path)
class GenxModuleFinder(MetaPathFinder):

    def find_spec(self, fullname, path=None, target=None):
        if fullname.startswith('genx'):
            return None
        itms=fullname.split('.')
        fname=os.path.join(os.path.dirname(os.path.abspath(__file__)), *itms)+'.py'
        pname=os.path.join(os.path.dirname(os.path.abspath(__file__)), *itms, '__init__.py')
        if os.path.exists(fname):
            m=import_module('genx.'+fullname)
            sys.modules[fullname]=m
            return spec_from_file_location(fullname, fname)
        if os.path.exists(pname):
            m=import_module('genx.'+fullname)
            sys.modules[fullname]=m
            return spec_from_file_location(fullname, pname)

sys.meta_path.insert(0, GenxModuleFinder())



#
# spec.py (c) Stuart B. Wilkins 2008
#
# $Id$
# $HeadURL$
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Part of the "pyspec" package
#

"""Python module for handling of spec data files (pyspec).

This module defines classes, objects and routines for reading SPEC datafiles
into python classes. There are three main classes for the data, SpecDataFile,
SpecScan and SpecData and are arranged as follows:

SpecDataFile, contains a single data file. The datafile is indexed upon
initializing the class to allow for fast searching of large data files.

To retrieve a scan from the datafile, an item is requested from the object. For
example;

    sd = SpecDataFile('data.01')
    scan = sd[100]

The construct sd[100] will return scan 100 as a SpecScan object. By passing a range
of integers, for example;

    scans = sd[[100, 101]]

these data will be concatenated into a single object. This is useful, for example,
when two scans are used to take the data (one wide, one close) and you need to fit
the data.

The SpecScan object contains members for all motors and counters for the scan in
question. These can be accessed in two main ways. For example, to obtain the
detector counts for the scan as an array;

   Det = sd[1000].Detector

or

   scan = sd[1000]
   Det = scan.Detector

can be used. For situations where you need machine adjustable code, the construct

   Det = scan.values['Detector']

can be used, where the member 'values' is a python dictionary of all scan variables.
In addition to the counters and motors, values defined in the header are also included.
See the SpecScan class documentation for further information.

Finally, the SpecScan class defines some helper routines for plotting, fitting etc.
which can be used to help in data analysis. See the documentation on the SpecScan class
for more information.

"""

import numpy

from pylab import *
from matplotlib import pylab
# from fit import fit
# from ccd.transformations import FileProcessor
from copy import copy
from genx.core.custom_logging import iprint

__version__="$Revision$"
__author__="Stuart B. Wilkins <swilkins@bnl.gov>"
__date__="$LastChangedDate$"
__id__="$Id$"
__verbose__=0x01

NotInFile="Scan is not in Datafile"

class SpecExtension:
    """Class to define extensions to SpecDataFile"""

    def __init__(self):
        return

    def getName(self):
        """Return string of name of extension"""
        return "Dummy"

    def initSpec(self, object):
        """Initialize SpecDataFile class"""
        return

    def initSpecScan(self, object):
        """Initialize SpecScan class"""
        return

    def parseSpecHeader(self, object, line):
        """Parse a line of a spec header and modify object with results"""
        return

    def parseSpecScanHeader(self, object, line):
        """Parse a line of a spec scan header and modify object with results"""
        return

    def postProcessSpecHeader(self, object):
        """Post process spec header"""
        return

    def postProcessSpecScanHeader(self, object):
        """Post process spec scan header"""
        return

    def concatenateSpecScan(self, object, a):
        return

# Now try to load default extensions
DefaultUserExtensions=[]
try:
    # noinspection PyUnresolvedReferences
    from ccd.specext import CCDSpecExtension

    DefaultUserExtensions.append(CCDSpecExtension())
except:
    pass

def removeIllegals(key):
    """Remove illegal character from string"""
    illegal=['/', ' ']
    for j in illegal:
        key=key.replace(j, '')
    if not key[0].isalpha():
        key="X"+key

    return key

def splitSpecString(ips):
    """Split a spec string which is formated with two spaces"""
    ops=[]
    for o in ips.split('  '):
        if o!='':
            ops.append(o.strip())

    return ops

class SpecDataFile:
    """ DataFile class for handling spec data files"""

    def __init__(self, fn, userext=None, **kwargs):
        """Initialize SpecDataFile

        Parameters
        ----------
        fn : string
             Filename of spec file to open
        userext : list
             A list of user extension classes to run on data

        Returns a SpecDataFile object

        """
        if userext is None:
            userext=[]
        self.filename=fn
        self.mode='concat'  # Set the default to concatenate multiple files

        # User extensions for adding functionality
        self.userExtensions=userext
        self.userExtensions=self.userExtensions+DefaultUserExtensions

        if __verbose__:
            for ext in self.userExtensions:
                iprint("**** Defining extension %s"%ext.getName())

        for ext in self.userExtensions:
            ext.initSpec(self)

        for arg in kwargs:
            # Set any keyword args into the base class
            setattr(self, arg, kwargs[arg])

        self._loadSpecFile()
        return

    def _loadSpecFile(self):
        if __verbose__:
            iprint("**** Opening specfile %s."%self.filename)

        self.index()
        self.readHeader()
        iprint(self.getStats())

        self.scandata={}
        return

    def __getstate__(self):
        # Called to pickle class
        mydict=copy(self.__dict__)
        mydict['file']=None
        return mydict

    def reset(self):
        """Reset the specfile class

        This routine resets the SpecDataFile, as if no scans
        had been read."""

        self.scandata={}
        if __verbose__:
            iprint("**** SpecFile reset (all scans removed)")

    def reload(self):
        """Reload the data file

        This routine reloads (reindexes) the datafile"""

        if __verbose__:
            iprint("**** Reloading SpecFile")
        self._loadSpecFile()

    def setMode(self, mode='concatenate'):
        """Set the modee to deal with multiple scans

        mode : string
           If mode is 'concatenate' then concatenate scans together.
           If mode is 'bin' then bin (numerically the scans together.
        """

        if mode=='concatenate':
            self.mode='concat'
            iprint("**** Multiple scans will be concatenated.")
            return
        elif mode=='bin':
            self.mode='bin'
            iprint("**** Multiple scans will be binned.")
            return
        else:
            raise Exception("Unknown mode %s"%mode)

        return

    def readHeader(self):
        """Read the spec header from the datafile.

        Currently supported header items:
                '#O'    (Motor positions)
        """

        self.file=open(self.filename, 'r')

        if __verbose__:
            iprint("---- Reading Header.")

        self.motors=[]
        self.file.seek(0, 0)
        line=self.file.readline()
        while line[0:2]!="#S":
            if line[0:2]=="#O":
                self.motors=self.motors+splitSpecString(line[4:])
            else:
                # Run user extensions
                for ext in self.userExtensions:
                    ext.parseSpecHeader(self, line)

            line=self.file.readline()

        for ext in self.userExtensions:
            ext.postProcessSpecHeader(self)

        self.file.close()
        return

    def index(self):
        """Index the datafile

        This routine indexes and sorts the byte-offests for
        all the scans (Lines beginning with '#S')

        """
        self.file=open(self.filename, 'r')

        if __verbose__:
            iprint("---- Indexing scan :       ", end=' ')
            sys.stdout.flush()
            sys.stderr.flush()

        self.file.seek(0, 0)
        self.findex={}
        self.scan_commands={}

        pos=self.file.tell()
        line=self.file.readline()
        while line!="":
            if line[0:2]=="#S":
                a=line.split()
                s=int(a[1])
                if (s%5)==0:
                    iprint("\b\b\b\b\b\b\b%5d "%s, end=' ')
                    sys.stdout.flush()
                self.findex[s]=pos
                self.scan_commands[s]=' '.join(a[2:])
            pos=self.file.tell()
            line=self.file.readline()
        iprint("\b\b\b\b\b\b\bDONE  ")

        self.file.close()
        return

    def getStats(self, head="---- "):
        """ Returns string with statistics on specfile.

        Parameters
        ----------

        head : string
           append string head to status text

        """
        string=""
        string=string+head+"Specfile contains %d scans\n"%len(self.findex)
        string=string+head+"Start scan = %d\n"%min(self.findex.keys())
        string=string+head+"End   scan = %d\n"%max(self.findex.keys())

        return string

    def _moveto(self, item):
        """Move to a location in the datafile for scan"""

        if item in self.findex:
            self.file.seek(self.findex[item])
        else:
            # Try re-indexing the file here.
            if __verbose__:
                iprint("**** Re-indexing scan file\n")
            self.index()
            if item in self.findex:
                self.file.seek(self.findex[item])
            else:
                raise Exception("Scan %s is not in datafile ....."%item)

    def __getitem__(self, item):
        """Convinience routine to use [] to get scan"""
        if item<0:
            # Indexes the scan backwards from the highest number.
            items=list(self.findex.keys())
            items.sort()
            return self.getScan(items[item], setkeys=True)
        else:
            # Otherwise behave as normal
            return self.getScan(item, setkeys=True)

    def getAll(self, *args, **kwargs):
        """Read all scans into the object"""
        for s in list(self.findex.keys()):
            self.getScan(s, *args, **kwargs)

    def getScan(self, item, mask=None, setkeys=True, persistent=True,
                reread=False, **kwargs):
        """Get a scan from the data file

        This routine gets a scan from the data file and loads it into the
        list of SpecData instances.

        setting 'setkeys = True' will set attributes to the
        specdata item of all the motors and counters

        The mask can be used to detete bad points. This should be a list of datum 
        (data point numbers) which should be excluded. If the "item" is a list of
        scans (that will be concatenated or binned together) this should be a list
        containing either None for no removal of data points or a list of points to
        remove

        Returns the ScanData object corresponding to the scan requested.

        """

        if type(item)==int:
            items=(item,)
        elif type(item)==float:
            items=(int(item),)
        elif type(item)==list:
            items=tuple(item)
        elif type(item)==tuple:
            items=item
        elif type(item)==numpy.ndarray:
            items=item.tolist()
        else:
            raise Exception("item can only be <int>, <float>, <list>, <array> or <tuple>")

        if mask is None:
            mask=[None for i in items]

        if len(mask)!=len(items):
            raise Exception("The mask list should be the same size as the items list")

        self.file=open(self.filename, 'r')
        rval=[]
        n=0
        for i, m in zip(items, mask):
            if __verbose__:
                iprint("**** Reading scan/item %s"%i)
            if ((i in self.scandata) is False) or (reread is True):
                self._moveto(i)
                self.scandata[i]=SpecScan(self, i, setkeys, mask=m, **kwargs)

            rval.append(self.scandata[i])

        self.file.close()

        if len(rval)>1:
            newscan=rval[0]  # deepcopy(rval[0])
            self.scandata.pop(newscan.scan)
            scans=[newscan.scan]
            for i in range(len(rval)-1):
                if self.mode=='concat':
                    newscan.concatenate(rval[i+1])
                elif self.mode=='bin':
                    newscan.bin(rval[i+1], binbreak='Seconds')
                else:
                    raise Exception("Unknown mode to deal with multiple scans.")
                scans.append(rval[i+1].scan)
            newscan.scan=scans
            newscan.scan_command='Concatenated'
            rval=[newscan]

        return rval[0]

    def _getLine(self):
        """Read line from datafile"""

        line=self.file.readline()
        if __verbose__ & 0x10:
            iprint("xxxx %s"%line.strip())
        return line

class SpecScan:
    """Class defining a SPEC scan

    This class defines a single spec scan, or a collection of scans either binned or concatenated.
    The class has members for the data and variables contained in the scan. If the optional 'setkeys'
    is defined (see __init__) then the motor positions and variables will be set as class members. For
    example, in a typical scan the motor position TTH can be accessed as a class member by SpecScan.TTH

    This object has some standard members:

    header     : [string]  Header of spec scan
    values     : [dict]    All scan data variables (Motors and Counters)
    data       : [array]   All scan data (cols = variables, rows = datum)
    comments   : [string]  Comments inserted in scan
    scandate   : [time]    Date of start of scan

    There are a number of special members. Wherever possible the following members are added to
    the class:

    Qvec       : [array]   Q Vector
    alphabeta  : [array]   array([alpha, beta])
    wavelength : [float]   incident wavelength
    omega      : [float]   omega ((TTH / 2) - TH)
    azimuth    : [float]   azimuthal angle

    """

    def __init__(self, specfile, item, setkeys=True, mask=None, **kwargs):
        """Read scan data from SpecFile

        Initialize the SpecScan class from a SpecData instance.

        Parameters
        ----------
        specfile       : [SpecFile] Instance of a spec data file.
        item           : [item]     Scan item.
        setkeys        : [bool]     If true set items of the class
                                    from all variables.
        """
        from time import strptime

        # Keep track of the datafile

        self.datafile=specfile
        self.scandata=SpecData()
        self.scanplot=None
        self.scanplotCCD=None
        self.setkeys=setkeys

        # Define the SIXC angles

        self.sixcAngleNames=['Delta', 'Theta', 'Chi', 'Phi', 'Mu', 'Gamma']
        self.fourcAngleNames=['TwoTheta', 'Theta', 'Chi', 'Phi']

        line=specfile._getLine()

        if __verbose__:
            iprint("---- %s"%line.strip())

        sline=line.strip().split()

        self.scan=int(sline[1])
        self.scan_type=sline[2]
        self.scan_command=' '.join(sline[2:])

        self.header=line
        self.comments=""

        x=0
        self.values={}
        self.data=array([])

        self.UB=eye(3)

        line=specfile._getLine()
        self.header=self.header+line

        # Run the init for any user extensions
        for ext in specfile.userExtensions:
            ext.initSpecScan(self)

        # Finally overide any assigments with values passed as keyword arguments
        for arg in kwargs:
            setattr(self, arg, kwargs[arg])

        #
        # Read the spec header and place the data into this class
        #

        while (line[0:2]!="#L") & (line!=""):
            if line[0:2]=="#P":
                # Motor positions
                pos=line.strip().split()
                for i in range(1, len(pos)):
                    self.scandata.setValue(removeIllegals(specfile.motors[x]), array([float(pos[i])]))
                    x+=1

            elif line[0:2]=="#C":
                # Comments
                self.comments=self.comments+line

            elif line[0:2]=="#D":
                try:
                    self.scandate=strptime(line[2:].strip())
                except:
                    self.scandate=None
            elif line[0:3]=="#G4":
                try:
                    pos=line[3:].strip().split()
                    self.Qvec=array([float(pos[0]), float(pos[1]), float(pos[2])])
                    self.alphabeta=array([float(pos[4]), float(pos[5])])
                    self.wavelength=float(pos[3])
                    self.energy=12398.4/self.wavelength
                    self.omega=float(pos[6])
                    self.azimuth=float(pos[7])
                except:
                    iprint("**** Unable to read geometry information (G4)")
            elif line[0:3]=="#G1":
                try:
                    pos=line[3:].strip().split()
                    pos=array(list(map(float, pos)))

                    self.Lattice=pos[0:6]
                    self.RLattice=pos[6:12]
                    self.or0=pos[12:15]
                    self.or1=pos[15:18]
                    sa=pos[18:-2].reshape(2, -1)
                    self.or0Angles=sa[0, :]
                    self.or1Angles=sa[1, :]
                    self.or0Lambda=pos[-2]
                    self.or1Lambda=pos[-1]
                except:
                    iprint("**** Unable to read geometry information (G1)")

            elif line[0:3]=="#G3":
                try:
                    pos=line[3:].strip().split()
                    pos=array(list(map(float, pos)))

                    self.UB=pos.reshape(-1, 3)
                except:
                    iprint("**** Unable to read UB matrix (G3)")
            else:
                # Try using the user extensions to parse the lines
                for ext in specfile.userExtensions:
                    ext.parseSpecScanHeader(self, line)

            line=specfile._getLine()
            self.header=self.header+line

        if line[0:2]=="#L":
            # Comment line just before data
            self.cols=splitSpecString(line[3:])
            iprint("---- %s"%line.strip())

        line=specfile._getLine()
        self.header=self.header+line

        iprint("---- %s"%line.strip())

        while (line[0:2]!="#S") & (line!="") & (line[0:4]!="# CM"):
            if line[0]!="#":
                datum=array([])
                d=line.strip().split()
                if len(d)!=0:
                    for i in range(len(d)):
                        v=array([float(d[i])])
                        datum=concatenate((datum, v), 1)

                    if self.data.size==0:
                        self.data=datum
                    else:
                        self.data=vstack((self.data, datum))

            elif line[0:2]=='#C':
                self.comments=self.comments+line
            else:
                self.header=self.header+line

            line=specfile._getLine()

        if self.data.ndim==1:
            self.data=numpy.array([self.data])

        if mask is not None:
            if __verbose__:
                iprint("---- Removing rows %s from data."%str(mask))
            self.data=np.delete(self.data, mask, axis=0)

        self.scanno=numpy.ones(self.data.shape[0], dtype=numpy.int)*self.scan
        self.scandatum=arange(self.data.shape[0])

        # Run the extension post processing scripts

        for ext in specfile.userExtensions:
            if __verbose__:
                iprint("---- Using extension %s"%ext.getName())
            ext.postProcessSpecScanHeader(self)

        # Now set the motors
        self._setcols()

        if __verbose__:
            iprint("---- Data is %i rows x %i cols."%(self.data.shape[0], self.data.shape[1]))

    def _setcols(self):
        if self.data.shape[0]>0:
            for i in range(len(self.cols)):
                if len(self.data.shape)==2:
                    self.scandata.setValue(removeIllegals(self.cols[i]), self.data[:, i])
                else:
                    self.scandata.setValue(removeIllegals(self.cols[i]), array([self.data[i]]))

            # Now set the variables into the scan class from the data

            if self.setkeys:
                for i in list(self.scandata.values.keys()):
                    if __verbose__ & 0x02:
                        iprint("oooo Setting variable %s"%i)
                    setattr(self, i, self.scandata.values[i])
                self.values=self.scandata.values

    def concatenate(self, a):
        # Could put check in here for cols matching ?!?

        self.header=self.header+a.header
        self.data=vstack((self.data, a.data))
        self.scanno=concatenate((self.scanno, a.scanno))
        self.scandatum=concatenate((self.scandatum, a.scandatum))

        for ext in self.datafile.userExtensions:
            ext.concatenateSpecScan(self, a)

        self._setcols()

    def bin(self, a, binbreak=None):
        """Bin the scans together adding the column values

        a is a SpecScan object of the file to bin.

        Note:
        This routine only bins the "data" portion of the scan. It returns the
        origional scans motors etc.

        """
        # First check if scans are the same.
        if self.cols!=a.cols:
            raise Exception("Scan column headers are not the same.")
        self.header=self.header+a.header
        if binbreak is not None:
            if binbreak in self.cols:
                flag=False
                for i in range(len(self.cols)):
                    if self.cols[i]==binbreak:
                        flag=True
                    if flag:
                        self.data[:, i]=self.data[:, i]+a.data[:, i]
            else:
                raise Exception("'%s' is not a column of the datafile."%binbreak)
        else:
            self.data=self.data+a.data
        self._setcols()

        return self

    def __getstate__(self):
        # Copy the dictionary and then return the copy
        return self.__dict__

    def getSIXCAngles(self):
        """This function returns the SIXC angles
        for the scan as a Nx6 array"""
        self.sixcAngles=zeros((self.data.shape[0], 6))
        for i, name in zip(list(range(6)), self.sixcAngleNames):
            v=self.scandata.get(name)
            if v.size==1:
                v=ones(self.data.shape[0])*v
            self.sixcAngles[:, i]=v

        return self.sixcAngles

    def plot(self, *args, **kwargs):
        """Plot the SpecScan using matplotlib"""

        if self.scanplot is None:
            self.scanplot=SpecPlot(self)

        self.scanplot.show(*args, **kwargs)
        return self.scanplot

    ## def plotCCD(self,  *args, **kwargs):
    ##     """Plot the SpecScan CCD images using matplotlib"""

    ##     if self.scanplotCCD is None:
    ##         self.scanplotCCD = SpecPlotCCD(self)

    ##     self.scanplotCCD.show(*args, **kwargs)
    ##     return self.scanplotCCD

    # def fit(self, funcs, quiet = False):
    #    """
    #    Overloaded function which calls the current scanplot
    #    with scanplot.fit(...)
    #    """
    #    if self.scanplot is None:
    #        raise Exception("You need to plot something before trying to fit it!")
    #    else:
    #        return self.scanplot.fit(funcs, quiet)

    def __str__(self):
        return self.show()

    def show(self, prefix="", nperline=4):
        """Return string of statistics on SpecScan"""
        p=""
        p=p+"Scan:\n\n"
        p=p+"\t%s\n\n"%self.scan
        p=p+"Datafile:\n\n"
        p=p+"\t%s\n\n"%self.datafile.file.name
        p=p+"Scan Command:\n\n"
        p=p+"\t%s\n\n"%self.scan_command
        p=p+"Scan Constants:\n\n"

        j=nperline
        typestoprint=[float, str, numpy.ndarray, int, numpy.float64]

        for d in self.__dict__:
            if d not in self.scandata.values:
                if typestoprint.count(type(getattr(self, d))):
                    p=p+"%-19s "%d
                    iprint(d, type(getattr(self, d)))
                    j-=1
                    if j==0:
                        p=p+"\n"
                        j=nperline

        p=p+"\n\n"
        p=p+self.scandata.show(prefix, nperline)
        return p

    def getYE(self, ycol=None, mcol=None):
        """Return an tuple of two arrays of y and e"""

        if type(ycol)==str:
            ycol=self.scan.cols.index(ycol)
        if type(mcol)==str:
            mcol=self.scan.cols.index(mcol)
        if ycol is None:
            ycol=-1
        if mcol is None:
            mcol=-2

        y=self.data[:, ycol]
        m=self.data[:, mcol]

        e=sqrt(y)/y
        y=y/m
        e=e*y

        return y, e

class SpecData:
    """Class defining the data contained in a scan"""

    def __init__(self):
        self.values={}

    def setValue(self, key, data, setdata=True):
        self.values[key]=data
        if __verbose__ & 0x20:
            iprint("oooo Setting key %s"%key)

    def get(self, key):
        if key in self.values:
            return self.values[key]
        else:
            return None

    def __str__(self):
        return self.show()

    def show(self, prefix="", nperline=6):
        """Return string of statistics on data (motors, scalars)"""

        j=nperline
        p=""
        p=p+prefix+"Motors:\n\n"
        p=p+prefix
        for i in list(self.values.keys()):
            if self.values[i].size==1:
                p=p+"%-19s "%i
                j-=1
                if j==0:
                    p=p+"\n"+prefix
                    j=nperline

        if j!=nperline:
            p=p+"\n"+prefix

        p=p+"\n"

        p=p+prefix+"\n"
        p=p+prefix+"Scan Variables:\n"
        p=p+prefix+"\n"
        j=nperline
        for i in list(self.values.keys()):
            if self.values[i].size>1:
                p=p+"%-19s "%i
                j-=1
                if j==0:
                    p=p+"\n"+prefix
                    j=nperline

        p=p+"\n"
        return p

class SpecPlot:
    def __init__(self, specscan):
        self.scan=specscan
        self.plt=None

    def show(self, xcol=None, ycol=None, mcol=None,
             norm=True, doplot=True, errors=True,
             fmt='ro', new=True,
             xint=200, yint=200,
             notitles=False, log=False, twodtype='contour'):
        """Plot and display the scan

        'xcol', 'ycol' and 'mcol' can be either numbers (negative count from end)
        or string values for the motor names

        for 2D plots 'xcol' can be a list or array e.g. [0, 1]

        'norm = False' will suppress normalization of the data
        'doplot = False' will suppress the plotting of the figure
        'errors = False' will suppress the plotting of errorbars
        'fmt' is the format string passed onto the plot command
        'new = True' will create a new figure
        'xint = 200' will set the x size of the 2D grid to 200 pts
        'yint = 200' will set the y size of the 2D grid to 200 pts
        'log = True' will set the y axis to logscale
        'twodtype = 'contout''
        """

        twod=False
        x2col=None

        if ycol is None:
            ycol=-1
        if mcol is None:
            mcol=-2

        if xcol is None:
            if self.scan.scan_type.strip()=='hklmesh':
                # Look at H K and L and see which varies the most
                # This is a cludge, surely there is a better way of
                # doing this!

                v=self.scan.data[:, 0:3].var(axis=0)
                v=vstack((arange(0, 3), v))
                v=v.transpose()
                v=v[argsort(v[:, 1], 0)]
                xcol=int(v[2, 0])
                x2col=int(v[1, 0])
                twod=True
            elif self.scan.scan_type.strip()=='mesh':
                x2col=1
                xcol=0
                twod=True
            else:
                x2col=None
                xcol=0

        if (type(xcol)==list) | (type(xcol)==ndarray):
            x2col=xcol[1]
            xcol=xcol[0]
        else:
            if (type(xcol)!=int) & (type(xcol)!=str):
                raise Exception("Illegal '%s' : xcol can only be 'int', 'list' or 'ndarray'."%type(xcol))

        if type(xcol)==str:
            xcol=self.scan.cols.index(xcol)
        if type(x2col)==str:
            x2col=self.scan.cols.index(x2col)
        if type(ycol)==str:
            # Not a number so it is probably a name
            ycol=self.scan.cols.index(ycol)
        if type(mcol)==str:
            mcol=self.scan.cols.index(mcol)

        if x2col is not None:
            twod=True

        self.plotx=self.scan.data[:, xcol]
        if x2col is not None:
            self.plotx=vstack((self.plotx, self.scan.data[:, x2col]))
            self.plotx=self.plotx.transpose()

        if __verbose__:
            iprint("**** Plotting scan %s (%s)"%(self.scan.scan, self.scan.scan_command))
            iprint("---- x  = %s"%self.scan.cols[xcol])
            if x2col is not None:
                iprint("---- x2 = %s"%self.scan.cols[x2col])

        if norm:
            self.ploty=self.scan.data[:, ycol]/self.scan.data[:, mcol]
            self.plote=self.ploty*(sqrt(self.scan.data[:, ycol])/self.scan.data[:, mcol])
            yl="%s / %s"%(self.scan.cols[ycol], self.scan.cols[mcol])
            if __verbose__:
                iprint("---- y  = %s / %s"%(self.scan.cols[ycol], self.scan.cols[mcol]))
        else:
            self.ploty=self.scan.data[:, ycol]
            self.plote=sqrt(self.ploty)
            yl=self.scan.cols[ycol]
            if type(errors)==type(''):
                self.plote=self.scan.data[:, self.scan.cols.index(errors)]
            if __verbose__:
                iprint("---- y  = %s"%self.scan.cols[ycol])
        if log:
            plotit=semilogy
        else:
            plotit=plot

        if twod:
            # Add check for matplotlib version
            pass

        # Plot the data
        if doplot:
            from scipy.interpolate import griddata
            if new:
                pylab.figure()

            pylab.hold(False)

            if twod:
                xi=linspace(min(self.plotx[:, 0]), max(self.plotx[:, 0]), xint)
                yi=linspace(min(self.plotx[:, 1]), max(self.plotx[:, 1]), yint)
                iprint("---- xint = %d"%xint)
                iprint("---- yint = %d"%yint)
                if log:
                    zi=griddata(self.plotx[:, 0], self.plotx[:, 1], numpy.log10(self.ploty), xi, yi)
                else:
                    zi=griddata(self.plotx[:, 0], self.plotx[:, 1], self.ploty, xi, yi)
                if twodtype=='pcolor':
                    pcolor(xi, yi, zi)
                else:
                    contour(xi, yi, zi)
                if log:
                    colorbar(format=FormatStrFormatter('$10^{%d}$'))
                else:
                    colorbar()

                xlim([min(self.plotx[:, 0]), max(self.plotx[:, 0])])
                ylim([min(self.plotx[:, 1]), max(self.plotx[:, 1])])

                if not notitles:
                    xlabel(self.scan.cols[xcol])
                    ylabel(self.scan.cols[x2col])
            else:
                dy=max(self.ploty)-min(self.ploty)
                mde=mean(self.plote)
                if (mde>dy) | isnan(mde):
                    # errors = False
                    iprint("---- Errorbars disabled due to large errors")
                if errors:
                    pylab.hold(True)  # Needed for errorbar plots
                    if log:
                        gca().set_yscale('log')
                    if fmt[0] in ['b', 'g', 'r', 'c', 'm', 'y', 'k', 'w']:
                        col=fmt[0]
                    else:
                        col='r'
                    self.plt=errorbar(self.plotx, self.ploty, self.plote,
                                      ecolor=col, fmt=fmt)

                else:
                    self.plt=plot(self.plotx, self.ploty, fmt)

                grid(True)

                if not notitles:
                    xlabel(self.scan.cols[xcol])
                    ylabel(yl)

                xlim((min(self.plotx), max(self.plotx)))

            if not notitles:
                title("%s %s\n%s"%(self.scan.datafile.filename, self.scan.scan.__repr__(), self.scan.scan_command))
            self.plotted=True
        else:
            self.plotted=False

    # def fit(self, funcs, quiet = False):
    #
    #    f = fit(x = self.plotx, y = self.ploty, funcs = funcs)
    #    plsq = f.go()
    #
    #    # Generate a new values of x (500) to make plotted
    #    # functions look good!

    ## step = ( self.plotx.max() - self.plotx.min() ) / 500
    ## x = arange(self.plotx.min(), self.plotx.max(), step)

    ## self.fitx = x
    ## self.fity = f.evalfunc(x = x)

    ## if self.plotted:
    ##     hold(True)
    ##     plot(self.fit,self.fity, 'b-')

    ## return plsq

## class SpecPlotCCD():
##     """This class defines routines which plot a specfile CCD data"""

##     def __init__(self, specscan):
##         """Initialize class from specscan"""
##         self.scan = specscan
##         self.plt = None

##     def show(self, imagen = None, size = (4, 3), dark = True,
##              log = False, limits = None):
##         """Plot out CCD images.

##         imagen : numbers of images (data points) to plot as a list.
##         size   : grid to show images on"""

##         if imagen is None:
##             imagen = range(self.scan.data.shape[0])

##         images = FileProcessor(spec = self.scan)
##         images.process(dark)

##         if len(imagen) == 1:
##             f = figure()
##             ax1 = axes([0.35, 0.35, 0.6, 0.6])
##             ax2 = axes([0.35, 0.1, 0.6, 0.1])
##             ax3 = axes([0.1, 0.35, 0.1, 0.6])

##             data = images.GetImage(imagen[0])
##             if log:
##                 data = numpy.ma.array(data, mask = data <= 0)
##                 data = log10(data)

##             if limits is not None:
##                 dmin, dmax = setImageRange(data, limits)
##             else:
##                 dmin = data.min()
##                 dmax = data.max()

##             ims = ax1.imshow(data, vmin = dmin, vmax = dmax, cmap = cm.jet)
##             print ims.figbox
##             ax1.set_aspect(1./ax1.get_data_ratio())
##             ax2.plot(arange(data.shape[1]), data.sum(0))
##             ax2.set_xlim([0, data.shape[1]])
##             ax3.plot(data.sum(1), arange(data.shape[0]))
##             ax3.set_ylim([0, data.shape[0]])
##             #ax3.plot(arange(data.shape[0]),data.sum(1))
##             #ax2.set_ylim([dmin, dmax])
##         else:

##             x = 0
##             for i in imagen:
##                 if x == 0:
##                     f = figure()
##                 subplot(size[0], size[1], x)

##                 data = images.GetImage(i)
##                 if log:
##                     data = numpy.ma.array(data, mask = data <= 0)
##                     data = log10(data)

##                 if limits is not None:
##                     dmin, dmax = setImageRange(data, limits)
##                 else:
##                     dmin = data.min()
##                     dmax = data.max()

##                 imshow(data, vmin = dmin, vmax = dmax, cmap = cm.jet)
##                 title("%d.%d" % (self.scan.scan,i))
##                 x += 1
##                 if x == (size[0] * size[1]):
##                     x = 0
